import copy
import logging
import datasets


logger = logging.getLogger(__name__)


class Cell:
    """
    Table cell
    """

    def __init__(
        self,
        value=None,
        idx=None,
        colspan=1,
        rowspan=1,
        is_highlighted=False,
        is_col_header=False,
        is_row_header=False,
        is_dummy=False,
        main_cell=None,
    ):
        self.idx = idx
        self.value = value
        self.colspan = colspan
        self.rowspan = rowspan
        self.is_highlighted = is_highlighted
        self.is_col_header = is_col_header
        self.is_row_header = is_row_header
        self.is_dummy = is_dummy
        self.main_cell = main_cell

    @property
    def is_header(self):
        return self.is_col_header or self.is_row_header

    def serializable_props(self):

        ser_props = {}
        ser_props["idx"] = self.idx
        ser_props["value"] = str(self.value)
        ser_props["is_row_header"] = self.is_row_header
        ser_props["is_col_header"] = self.is_col_header
        return ser_props

    def __repr__(self):
        return str(vars(self))


class Table:
    """
    Table object
    """

    def __init__(self):
        self.props = {}
        self.cells = []
        self.outputs = {}
        self.url = None
        self.cell_idx = 0
        self.current_row = []
        self.cell_by_ids = {}
        self.default_question = ""
        self.type = "default" # default or custom
        self.custom_table_name = "" # for custom table
        self.txt_info = []
        self.pic_info = []
        self.is_linked = False
        self.html = "" # for multihiertt dataset

    def has_highlights(self):
        return any(cell.is_highlighted for row in self.cells for cell in row)

    def save_row(self):
        if self.current_row:
            self.cells.append(self.current_row)
            self.current_row = []

    def add_cell(self, cell):
        cell.idx = self.cell_idx
        self.current_row.append(cell)
        self.cell_by_ids[self.cell_idx] = cell
        self.cell_idx += 1

    def set_cell(self, i, j, c):
        self.cells[i][j] = c

    def get_cell(self, i, j):
        try:
            return self.cells[i][j]
        except:
            return None

    def get_cell_by_id(self, idx):
        return self.cell_by_ids[idx]

    def get_flat_cells(self, highlighted_only=False):
        return [x for row in self.cells for x in row if (x.is_highlighted or not highlighted_only)]

    def get_highlighted_cells(self):
        return self.get_cells(highlighted_only=True)

    def get_cells(self, highlighted_only=False):
        if highlighted_only:
            cells = []
            for row in self.cells:
                row_cells = [c for c in row if c.is_highlighted]
                if row_cells:
                    cells.append(row_cells)
            return cells
        else:
            return self.cells

    def get_row_headers(self, row_idx, column_idx):
        try:
            headers = [c for c in self.get_cells()[row_idx][:column_idx] if c.is_row_header]
            return headers

        except Exception as e:
            logger.exception(e)

    def get_col_headers(self, row_idx, column_idx):
        try:
            headers = []
            for i, row in enumerate(self.get_cells()):
                if i == row_idx:
                    return headers

                if len(row) > column_idx and row[column_idx].is_col_header:
                    headers.append(row[column_idx])

            return headers

        except Exception as e:
            logger.exception(e)

    def __repr__(self):
        return str(self.__dict__)



class TabularDataset:
    """
    Base class for the datasets
    """

    def __init__(self, path):
        self.splits = ["train", "dev", "test"]
        self.data = {split: [] for split in self.splits}
        self.tables = {split: {} for split in self.splits}
        self.path = path
        self.dataset_info = {}
        self.name = None

    def load(self, split):
        """
        Load the dataset. Path can be specified for loading from a directory
        or omitted if the dataset is loaded from HF.
        """
        raise NotImplementedError



    def get_data(self, split, table_idx):
        return self.data[split][table_idx]

    def set_table(self, split, table_idx, table):
        self.tables[split][table_idx] = table

    def get_example_count(self, split):
        return len(self.data[split])

    def has_split(self, split):
        return bool(self.data[split])

    def get_table(self, split, table_idx, edited_cells=None):
        table = self.tables[split].get(table_idx)
        if edited_cells:
            table_modif = copy.deepcopy(table)
            for cell_id, val in edited_cells.items():
                cell = table_modif.get_cell_by_id(int(cell_id))
                cell.value = val
            table = table_modif
        return table

    def prepare_table(self, entry):
        return NotImplementedError

    def get_info(self):
        return self.dataset_info

class HFTabularDataset(TabularDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, path=None, **kwargs)

        self.split_mapping = {"train": "train", "dev": "validation", "test": "test"}
        self.dataset_info = {}


    def _load_split(self, split):
        return NotImplementedError

    def load(self, split=None):
        if split is None:
            for split in self.split_mapping.keys():
                self._load_split(split)
        else:
            self._load_split(split)

    def save_to_disk(self, split, filepath):
        self.data[split].save_to_disk(filepath)
        logger.info(f"File {filepath} saved successfully")

    def load_from_disk(self, split, filepath):
        self.data[split] = datasets.load_dataset(filepath)



from dataclasses import dataclass, field
import random
from typing import Union, List

import numpy as np
import torch
from transformers import HfArgumentParser

@dataclass
class RunnningArguments:
    data_path: str = field(
        metadata={"help": "The path of mmqa dataset."}
    )
    caption_file_name: str=field(
        default='mmqa_captions_llava.json',
        metadata={"help": "Caption file name"}
    )
    split: str=field(
        default='validation',
        metadata= {
            "help": "run on which set, choices = ['train', 'validation', 'test']"
        }
    )
    retriever_files_path: str=field(
        default="utils/retriever_files/",
        metadata={
            "help": "TYPE_FILE: question_type_\{SPLIT\}.json"
                    "QIM_FILE: question_image_match_scores_\{SPLIT\}.json"
                    "QPM_FILE: question_passage_match_scores_\{SPLIT\}.json"
        }
    )
    top_n: int=field(
        default=3,
        metadata={
            "help": "use top_n retrieved files"
        }
    )
    template_path: str=field(
        default="mmhqa_icl/templates",
        metadata= {
            "help": "Path to the directory of templates"
        }
    )
    prompt_file: str=field(
        default="prompt.json",
        metadata={
            "help": "Name of the config file of icl"
        }
    )
    api_keys_file: str=field(
        default=None,
        metadata= {
            "help": "Path to the api_keys file"
        }
    )
    save_dir: str=field(
        default='results/',
        metadata= {
            "help": "Path to save directory"
        }
    )
    save_file_name: str=field(
        default=None,
    )

    # Multiprocess options
    n_processes: int=field(
        default=1,
    )

    # Generation options
    seed: int=field(
        default=42,
    )
    resume: bool=field(default=False, metadata={"help": "whether to resume from last result file"})
    oracle_classifier: bool=field(default=False, metadata={"help": "whether to use oracle classifier"})
    oracle_retriever: bool=field(default=False, metadata={"help": "whether to use oracle retriever"})

    # LLM options
    n_parallel_prompts: int=field(default=1)
    max_api_total_tokens: int=field(default=4001)
    temperature: float=field(default=0.4, metadata={"help": "temperature param of LLM"})
    engine: str=field(
        default="text-davinci-003",
        metadata={"help": "choices= ['text-davinci-003']"}
    )
    top_p: float=field(
        default=1.0
    )
    stop_tokens: str=field(
        default='\n\n\n',
        metadata={
            "help": 'Split stop tokens by ||'
        }
    )
    # debug options
    verbose: bool=field(
        default=False,
        metadata={
            "help": "debug options"
        }
    )

def get_running_args():
    return HfArgumentParser(
        RunnningArguments
    ).parse_args_into_dataclasses()[0]


import datasets
import os
import json


import logging
import sys
import math
import torch


def gelu(x):
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))

def create_logger(name, silent=False, to_disk=True, log_file=None):
    """Logger wrapper
    """
    # setup logger
    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG)
    log.propagate = False
    formatter = logging.Formatter(fmt='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S')
    if not silent:
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        log.addHandler(ch)
    if to_disk:
        log_file = log_file if log_file is not None else strftime("%Y-%m-%d-%H-%M-%S.log", gmtime())
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        log.addHandler(fh)
    return log


def load_json(path):
    with open(path, 'r') as f:
        data = json.load(f)
    return data


# def load_data(path_dict):
#     if 'train' in path_dict:
#         train_data = load_json(path_dict['train'])
#     if 'dev' in path_dict:
#         dev_data = load_json(path_dict['dev'])
#     if 'test' in path_dict:
#         test_data = load_json(path_dict['test'])
#     return {'train':train_data, 'dev':dev_data, 'test':test_data}







import math
import torch
import torch.nn as nn


def gelu(x):
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))


class FFNLayer(nn.Module):
    def __init__(self, input_dim, intermediate_dim, output_dim, dropout, layer_norm):
        super(FFNLayer, self).__init__()
        self.fc1 = nn.Linear(input_dim, intermediate_dim)
        self.layer_norm = nn.LayerNorm(intermediate_dim) if layer_norm else None
        self.dropout_func = nn.Dropout(dropout)
        self.fc2 = nn.Linear(intermediate_dim, output_dim)

    def forward(self, inputs):
        x = self.fc1(self.dropout_func(inputs))
        x = gelu(x)
        if self.layer_norm:
            x = self.layer_norm(x)
        return self.fc2(x)


def _get_loss_fn_and_output_dim(loss_fn_type):
    if loss_fn_type == "CE":
        return nn.CrossEntropyLoss(), 2
    elif loss_fn_type == "BCE":
        return nn.BCELoss(), 1
    else:
        raise ValueError(f'Unsupported loss_fn: {loss_fn_type}')


class Retriever(nn.Module):
    def __init__(self, encoder=None, loss_fn_type="CE", input_dim=1024, intermediate_dim=2048, dropout=0.1, layer_norm=True):
        super(Retriever, self).__init__()
        self.encoder = encoder
        self.loss_fn_type = loss_fn_type
        self.loss_fn, self.output_dim = _get_loss_fn_and_output_dim(loss_fn_type)
        self.classifier = FFNLayer(input_dim=input_dim, intermediate_dim=intermediate_dim,
                                   output_dim=self.output_dim, dropout=dropout, layer_norm=layer_norm)
        self.loss = None

    def forward(self, inputs):
        encoder_out = self.encoder(input_ids=inputs["input_ids"], attention_mask=inputs["attention_mask"])
        if self.loss_fn_type == "CE":
            output = self.classifier(encoder_out.last_hidden_state[:, 0, :])
        elif self.loss_fn_type == "BCE":
            output = self.classifier(encoder_out.last_hidden_state[:, 0, :])
            output = output.flatten().sigmoid()
            inputs["labels"] = inputs["labels"].flatten().float()
        else:
            raise NotImplementedError

        if inputs["labels"] is not None:
            self.loss = self.loss_fn(output, inputs["labels"])

        return output



import random
from typing import Dict, Tuple
import pandas as pd
import copy
import os
import json

class PromptBuilder(object):
    def __init__(self, args):
        self.args = args
        with open(os.path.join(args.data_path, args.caption_file_name), "r") as f:
            caption_map = json.load(f)
        self.caption_map = caption_map

    def _passage_prompt(self, passages, db_style_prompt=False):
        """
        Return the passage prompt.
        """
        if len(passages) == 0:
            return ""
        if not db_style_prompt:
            string = "Passages: \n"
            for passage_idx in range(len(passages['id'])):
                string += f"\"{passages['title'][passage_idx]}\"" + f" ({passages['text'][passage_idx]})" + '\n'
            string += '\n'
            return string
        else:
            passage_table_prompt = ""
            _header = []
            _rows = [[]]
            for passage in passages:
                # TODO: add passage prompt
                _header.append(passage['title'])
                _rows[0].append(passage['text'])

            return passage_table_prompt

    def _image_prompt(self, images, db_style_prompt=False):
        """
        Return the image prompt.
        """
        if len(images) == 0:
            return ""
        if not db_style_prompt:
            string = "Images: \n"
            for image_idx in range(len(images['id'])):
                string += f"\"{images['title'][image_idx]}\"" + f" ({self.caption_map[images['id'][image_idx]]})" + '\n'
            string += '\n'
            return string
        else:
            image_table_prompt = ""
            _header = []
            _rows = [[]]
            for image in images:
                # TODO: add image prompt
                _header.append(image['title'])
                _rows[0].append(image['caption'])
            return image_table_prompt

    def table2text_prompt(self, table, drop_row_id=False, ):
        _table = copy.deepcopy(table)
        header = _table['header'][0]
        rows = _table['rows'][0]
        table_title = _table['title'][0]
        if drop_row_id:
            if header[0] == "row_id":
                header = header[1:]
                rows = [_row[1:] for _row in rows]

        prompt_str = 'Table {}: \n'.format(table_title) if table_title else 'Table: \n'
        prompt_str += "\t".join(header) + "\n"
        prompt_str += '\n'.join(["\t".join([str(cell) if len(cell)>0 else '(null)' for cell in row]) for row in rows]) + "\n"
        return prompt_str

    def build_generate_prompt(
            self,
            qtype: Tuple,
            table: Dict = None,
            question: str = None,
            passages: Dict = None,
            images: Dict = None,
            cot: bool = False,
            supporting_context: Dict = None,
            **kwargs
    ):
        """
        Build the prompt of the generation sample.
        """
        generate_prompt = ""
        # task instruction
        if qtype == "image":
            generate_prompt += self._image_prompt(images=images)
            generate_prompt += f"Question: {question}\n"
        elif qtype == "text":
            generate_prompt += self._passage_prompt(passages=passages)
            generate_prompt += f"Question: {question}\n"
        elif qtype == "table":
            generate_prompt += self.table2text_prompt(table=table) + '\n'
            # generate_prompt += f"Question: {question}\n"
            generate_prompt += f"Question: {question}\n" # no type method
        else:
            generate_prompt += self._image_prompt(images=images)
            generate_prompt += self._passage_prompt(passages=passages)
            generate_prompt += self.table2text_prompt(table=table) + '\n'
            # generate_prompt += f"Question: {question}\n"
            generate_prompt += f"Question: {question}\n" # no type method

        if cot:
            if qtype in ['image']:
                generate_prompt += "Please answer the question step by step, you can make reasonable inference to give a definite answer.\n"
            else:
                generate_prompt += f"Please answer the question step by step.\n"
        else:
            generate_prompt += f"Answer: "

        return generate_prompt




import json
import os

ROOT_DIR = os.path.join(os.path.dirname(__file__), "../")

CAPTION_FILE = 'mmqa_captions.json'
# CAPTION_FILE = 'mmqa_captions_llava.json'
# CAPTION_FILE = 'mmqa_captions_llava_simple.json'

def get_caption_map(file_path=None):
    """
    Get the caption map.
    """
    if not file_path:
        file_path = os.path.join(ROOT_DIR, 'utils', CAPTION_FILE)

    with open(file_path, "r") as f:
        caption_map = json.load(f)
    return caption_map


def get_caption(id):
    """
    Get the caption of the picture by id.
    """
    with open(os.path.join(ROOT_DIR, 'utils', CAPTION_FILE), "r") as f:
        caption = json.load(f)
    if id in caption.keys():
        return caption[id]
    else:
        return ""



class Stack:
    def __init__(self):
        self.items = []
    def is_empty(self):
        return self.items == []
 
    def push(self, item):
        self.items.append(item)
 
    def pop(self):
        return self.items.pop()

    def peek(self):
        return self.items[len(self.items) - 1]
    def size(self):
        return len(self.items)
 
 
def infix_to_prefix(infix_expr):
    prec = dict()
    prec[")"] = 4
    prec["*"] = 3
    prec["/"] = 3
    prec["+"] = 2
    prec["-"] = 2
    prec["("] = 1
    prefix_expr = []
    s = Stack()
    for item in reversed(infix_expr.split()):
        if item not in prec.keys():
            prefix_expr.append(item)
        elif item == ')':
            s.push(item)
        elif item == '(':
            while s.peek() != ')':
                prefix_expr.append(s.pop())
            s.pop()
        else:
            while (not s.is_empty())\
                    and s.peek() != ')'\
                    and prec[s.peek()] > prec[item]:
                prefix_expr.append(s.pop())
                s.push(item)
            s.push(item)
    while not s.is_empty():
        prefix_expr.append(s.pop())
    prefix_expr.reverse()
    return ' '.join(prefix_expr)
 
 
def prefix_eval(prefix_expr):
    s = Stack()
    for item in reversed(prefix_expr.split()):
        if item not in '+-*/':
            s.push(item)
        else:
            op1 = float(s.pop())
            op2 = float(s.pop())
            result = do_match(item, op1, op2)
            s.push(result)
    return result
 
def do_match(op, op1, op2):
    if op == '+':
        return op1 + op2
    elif op == '-':
        return op1 - op2
    elif op == '*':
        return op1 * op2
    elif op == '/':
        return op1 / op2
    else:
        raise Exception('Error operation!')
    
    




from tatqa_metric import TaTQAEmAndF1

def test_em_and_f1():

    mode1_test_data = [
        ({'answer_type':'span', 'answer': ['here is, a test'], 'scale':''}, 'here is, a test', '', 1, 1),
        ({'answer_type': 'span', 'answer': ['here is, a test'], 'scale': ''}, 'here is, a test', '', 1, 1),
        ({'answer_type': 'span', 'answer': ['1234.1'], 'scale': 'million'}, '1234.1', 'thousand', 0, 0), # scale mismatch
        ({'answer_type': 'span', 'answer': ['1234.1'], 'scale': 'million'}, '123', 'thousand', 0, 0), # scale mismatch
        ({'answer_type': 'span', 'answer': ['12314.1'], 'scale': 'million'}, '12314.1', 'million', 1, 1),

        ({'answer_type': 'multi-span', 'answer': ['singapore', 'china', 'usa'], 'scale': ''}, ['singapore', 'china', 'usa'], '', 1, 1),
        ({'answer_type': 'multi-span', 'answer': ['singapore', 'china', 'usa'], 'scale': ''}, ['china', 'singapore', 'usa'], '', 1, 1),
        ({'answer_type': 'multi-span', 'answer': ['singapore', 'china', 'usa'], 'scale': ''}, ['china', 'singapore'], '',0, 0.8),

        ({'answer_type': 'arithmetic', 'answer': 123.2, 'scale': 'million'}, 123.2, '', 0, 0), # scale mismatch, f1 = 0
        ({'answer_type': 'arithmetic', 'answer': 123.2, 'scale': 'million'}, 123200000, '', 1, 1), #
        ({'answer_type': 'arithmetic', 'answer': 123.2, 'scale': 'million'}, 123.2, 'thousand', 0, 0), # scale mismatch
        ({'answer_type': 'arithmetic', 'answer': 123.2, 'scale': ''}, 123.2, '', 1, 1),
        ({'answer_type': 'arithmetic', 'answer': 123.22, 'scale': ''}, 123.2, '', 0, 0),
        ({'answer_type': 'arithmetic', 'answer': 123.2, 'scale': ''}, 123.2010, '', 1, 1),
        ({'answer_type': 'count', 'answer': 5, 'scale': ''}, 5, '', 1, 1),
        ({'answer_type': 'arithmetic', 'answer': 22.12, 'scale': 'percent'}, 0.2212, '', 1, 1),
        ({'answer_type': 'arithmetic', 'answer': 22.12, 'scale': 'percent'}, 0.22121, 'percent', 0, 0),
        ({'answer_type': 'arithmetic', 'answer': 22.12, 'scale': 'percent'}, 22.1231, '', 0, 0),
        ({'answer_type': 'arithmetic', 'answer': 22.12, 'scale': 'percent'}, 22.1231, 'percent', 1, 1),
        ({'answer_type': 'span', 'answer': [22.12], 'scale': 'million'}, '22.12', 'million', 1, 1),
        ({'answer_type': 'span', 'answer': [22.12], 'scale': 'million'}, '22.12', '', 0, 0),
        ({'answer_type': 'arithmetic', 'answer': 22.12, 'scale': 'million'}, 'test', '', 0, 0),
        ({'answer_type': 'arithmetic', 'answer': 22.12, 'scale': 'million'}, ["1","2"], '', 0, 0),# span is calcuated by word f1
        ({'answer_type': 'span', 'answer': [22.12], 'scale': 'percent'},"-22.12", '', 0, 0),
        ({'answer_type': 'span', 'answer': [22.12], 'scale': 'percent'},"22.12%", '', 1, 1),
        ({'answer_type': 'span', 'answer': [22.12], 'scale': ''}, "22.12%", '', 0, 0),
        ({'answer_type': 'span', 'answer': [22.12], 'scale': 'million'}, "$22.12", '', 0, 0),
        ({'answer_type': 'arithmetic', 'answer': 22.12, 'scale': 'million'}, "$22.12", '', 0, 0),
        ({'answer_type': 'span', 'answer': ["22.12"], 'scale': 'percent'}, ["-22.12"], '', 0, 0),
        ({'answer_type': 'span', 'answer': ['$1.0 million'], 'scale': ''}, ["['$1.0 million']"], '', 1, 1),

        ({'answer_type': 'span', 'answer': [22.12], 'scale': ''}, "$22.12", '', 1, 1),
        ({'answer_type': 'span', 'answer': [22.12], 'scale': 'percent'}, "22.12%", 'percent', 1, 1),
        ({'answer_type': 'count', 'answer': 5, 'scale': ''}, 'abcd 5', '1', 0, 0),

        ({'answer_type': 'multi-span', 'answer': ['$23,234', '$234.12'], 'scale': ''}, ['234.12', '23,234'], '',
         1, 1),
        ({'answer_type': 'multi-span', 'answer': ['$35,120', '$24,159'], 'scale': ''}, ['$24,159', '$35,120'], '', 1, 1),
        ({'answer_type': 'arithmetic', 'answer': ['34.12'], 'scale': 'percent'}, ['0.3412'], '', 1, 1),
        ({'answer_type': 'span', 'answer': [
            'wages and salaries, social security costs, pension and other costs and share-based payments, see note 10 of the Financial Statements'],
          'scale': ''},
         ['wages and salaries, social security costs, pension and other costs and share - based payments,'], '', 0,
         0.67),

    ]
    metrics = TaTQAEmAndF1()

    for ans, pred, pred_scale, em, f1 in mode1_test_data:
        metrics(ans, pred, pred_scale)
        pred_em, pred_f1, scale_score, op_score = metrics.get_overall_metric(reset=True)
        assert pred_em == em, f'mode2 - pred_em: {pred_em}, em:{em}, pred:{pred}, ans:{ans}'
        assert pred_f1 == f1, f'mode2 - pred_f1: {pred_f1}, f1:{f1}, pred:{pred}, ans:{ans}'


def test_one():
    mode_test_data = [
        ({'answer_type': 'arithmetic', 'answer': ['34.12%'], 'scale': 'percent'}, ['0.3412'], '', 1, 1),
        ({'answer_type': 'arithmetic', 'answer': ['34.12%'], 'scale': ''}, ['0.3412'], '', 1, 1),
    ]
    metrics = TaTQAEmAndF1()
    for ans, pred, pred_scale, em, f1 in mode_test_data:
        metrics(ans, pred, pred_scale)
        pred_em, pred_f1, scale_score, op_score = metrics.get_overall_metric(reset=True)
        assert pred_f1 == f1, f'mode2 - pred_f1: {pred_f1}, f1:{f1}, pred:{pred}, ans:{ans}'



import torch
from torch.utils.data import Dataset


class RetrieveDataset(Dataset):
    def __init__(self, data_list):
        self.data = data_list

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)


class RetrieveRowDataset(Dataset):
    def __init__(self, data_list):
        self.data = []
        self.positive_index = []
        idx = 0
        for one in data_list:
            for i in range(len(one["rows"])):
                sample = {"id": one["id"], "question": one["question"], "rows": one["rows"][i], "labels": []}
                sample["labels"].append(one["labels"][i])
                self.data.append(sample)
                if one["labels"][i]:
                    self.positive_index.append(idx)
                idx = idx + 1

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)

    def make_weights(self):
        weights = torch.ones(self.__len__()).float().cuda()
        weights[self.positive_index] = (self.__len__() - len(self.positive_index)) / len(self.positive_index)
        return weights



import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
from .utils import *
from transformers import AutoModel, AutoTokenizer, AutoConfig, Seq2SeqTrainer, TapexTokenizer, set_seed
from transformers import get_linear_schedule_with_warmup, AdamW
from tqdm import tqdm
import torch.nn.functional as F
import numpy as np
import argparse
from .model import Retriever
from .dataset import RetrievalDataset, ColumnRetrievalDataset
from typing import Dict, List, Tuple
from pathlib import Path



class RetrievalTrainer:
    def __init__(self, args):
        self.args = args
    
        Path(args.output_path).mkdir(parents=True, exist_ok=True)
        self.logger = create_logger("ReadModel", log_file=os.path.join(args.output_path, 'log.txt'))
        self.tokenizer = AutoTokenizer.from_pretrained(args.plm)

        if args.is_train:
            self.train_data = load_json(args.train_data_path)
            self.logger.info(f"train_data: {len(self.train_data)}")
        self.dev_data = load_json(args.dev_data_path)
        self.logger.info(f"dev_data: {len(self.dev_data)}")
        
        if args.mode == 'row':
            if args.is_train:
                self.train_dataset = RetrievalDataset(args, self.train_data)
            self.dev_dataset = RetrievalDataset(args, self.dev_data)   
        elif args.mode == 'column':
            if args.is_train:
                self.train_dataset = ColumnRetrievalDataset(args, self.train_data[:30])
            self.dev_dataset = ColumnRetrievalDataset(args, self.dev_data[:10])   
    
    # Dataset Collator
    def collate(self, data, **training_config):
        pad_id = self.tokenizer.pad_token_id
        bs = len(data)
        # 按bs为1去写
        data = data[0]
        
        
        input_ids = data['input_ids']
        max_input_length = max([len(item) for item in input_ids])
        if max_input_length > training_config['max_len']:
            max_input_length = training_config['max_len']
        
        input_ids_new = []
        for row_ids in input_ids:
            if len(row_ids) > max_input_length:
                row_ids = row_ids[:max_input_length]
            else:
                row_ids = row_ids + [pad_id] * (max_input_length - len(row_ids))
            input_ids_new.append(row_ids)
        
        input_ids = torch.tensor(input_ids_new)
        attention_mask = torch.where(input_ids==self.tokenizer.pad_token_id, 0, 1) #忽略填充pad影响
        if data['labels'].count(1) != 0:
            labels = torch.tensor(data['labels']) / data['labels'].count(1)
        return {"input_ids": input_ids.cuda(), "attention_mask": attention_mask.cuda(), "labels": labels.cuda(), "metadata":data}

    def train(self, **training_config):
        self.training_config = training_config
        
        train_loader = DataLoader(self.train_dataset, batch_size=training_config['train_bs'], collate_fn=lambda x: self.collate(x, **training_config))
        dev_loader = DataLoader(self.dev_dataset, batch_size=training_config['dev_bs'], collate_fn=lambda x: self.collate(x, **training_config))

        bert_model = AutoModel.from_pretrained(self.args.plm)
        model = Retriever(bert_model)
        device = torch.device("cuda")
        model.to(device)
        
        best_acc = -1
        self.optimizer = AdamW(model.parameters(), lr=training_config['lr'], eps=training_config['eps'])
        t_total = len(self.train_dataset) * training_config['epoch_num']
        self.scheduler = get_linear_schedule_with_warmup(self.optimizer, num_warmup_steps=0 * t_total, num_training_steps=t_total)

        for epoch in range(training_config['epoch_num']):
            self.logger.info(f"Training epoch: {epoch}")
            self.train_epoch(train_loader, model)
            self.logger.info("start eval....")
            acc = self.eval(model, dev_loader)
            self.logger.info(f"acc... {acc}")
            if acc > best_acc:
                best_acc = acc
                model_to_save = model.module if hasattr(model, "module") else model
                model_save_path = os.path.join(self.args.output_path, 'ckpt.pt')
                self.logger.info(f"saving model...to {model_save_path}")
                torch.save(model_to_save.state_dict(), model_save_path)

        
    def train_epoch(self, loader, model):
        model.train()
        averge_step = len(loader) // 12
        loss_sum, step = 0, 0
        for i, data in enumerate(tqdm(loader)):
            probs = model(data)
            loss_func = nn.BCEWithLogitsLoss(reduction='sum')
            loss = loss_func(probs, data['labels'].unsqueeze(-1))
            self.optimizer.zero_grad()
            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), 0.1)
            self.optimizer.step()
            self.scheduler.step()
            loss_sum += loss
            step += 1
            if i % averge_step == 0:
                self.logger.info("Training Loss [{0:.5f}]".format(loss_sum/step))
                loss_sum, step = 0, 0
    
    def eval(self, model, loader):
        model.eval()
        total, acc = 0, 0
        with torch.no_grad():
            for i, data in enumerate(tqdm(loader)):
                probs = model(data)
                predicts = torch.argmax(probs).cpu().item()
                gold_row = torch.where(data['labels']!=0)[0].cpu().tolist()

                total += 1
                if predicts in gold_row:
                    acc += 1
        print(f"Total: {total}")
        return acc / total
        
    



import time
import json
import argparse
import copy
import os

from typing import List
import platform
import multiprocessing

from utils import load_data_split, get_type
from utils.retriever import Retriever

ROOT_DIR = os.path.join(os.path.dirname(__file__), "../")


def work_eval(
        args,
        g_eids: List,
        dataset,
        retriever
):  
    g_dict = dict()
    total_gold_passages = 0
    total_gold_images = 0
    total_retrieved_passages = 0
    total_retrieved_images = 0
    gold_retrieved_passages = 0
    gold_retrieved_images = 0
    max_gold_passages = 0
    max_gold_images = 0
    gold_types = 0
    print(len(g_eids))
    for g_eid in g_eids:
        g_data_item = dataset[g_eid]
        g_dict[g_eid] = {
            'generations': dict(),
            'ori_data_item': copy.deepcopy(g_data_item)
        }
        table = g_data_item['table']
        new_passages, new_images, qtype = retriever.retrieve(g_data_item, g_eid, g_data_item['id'])
        # if qtype is None:
            # AssertionError("qtype is None")
            
        if (qtype is not None) and qtype == get_type(g_data_item['type']):
            gold_types += 1

        supc = g_data_item['supporting_context']
        gold_passage_id = [id for idx,id in enumerate(supc['doc_id']) if supc['doc_part'][idx] == 'text']
        gold_image_id   = [id for idx,id in enumerate(supc['doc_id']) if supc['doc_part'][idx] == 'image']

        gold_passage_id = list(set(gold_passage_id))
        gold_image_id = list(set(gold_image_id))
        

        if len(gold_passage_id) > max_gold_passages:
            max_gold_passages = len(gold_passage_id)
            # print(g_data_item['question'])

        if len(gold_image_id) > max_gold_images:
            max_gold_images = len(gold_image_id)
            # print(g_data_item['question'])

        if len(gold_passage_id) > 5:
            print(g_data_item['question'])
            
        total_gold_passages += len(gold_passage_id)
        total_gold_images += len(gold_image_id)
        total_retrieved_passages += len(new_passages['id'])
        total_retrieved_images += len(new_images['id'])
        gold_retrieved_passages += len(set(gold_passage_id) & set(new_passages['id']))
        gold_retrieved_images += len(set(gold_image_id) & set(new_images['id']))

    print(f"Type Accuracy:\t {gold_types / len(g_eids)}")
    print(f"Max gold passages:\t {max_gold_passages}")
    print(f"Max gold images:\t {max_gold_images}")
    print(f"Total gold passages:\t {total_gold_passages}")
    print(f"Total gold images:\t {total_gold_images}")
    print(f"Total retrieved passages:\t {total_retrieved_passages}")
    print(f"Total retrieved images:\t {total_retrieved_images}")
    print(f"Gold retrieved passages:\t {gold_retrieved_passages}")
    print(f"Gold retrieved images:\t {gold_retrieved_images}")
    # recall
    print(f"Passage recall:\t {gold_retrieved_passages / total_gold_passages}")
    print(f"Image recall:\t {gold_retrieved_images / total_gold_images}")
    # precision
    print(f"Passage precision:\t {gold_retrieved_passages / total_retrieved_passages}")
    print(f"Image precision:\t {gold_retrieved_images / total_retrieved_images}")
    # f1
    print(f"Passage f1:\t {2 * gold_retrieved_passages / (total_gold_passages + total_retrieved_passages)}")
    print(f"Image f1:\t {2 * gold_retrieved_images / (total_gold_images + total_retrieved_images)}")
    # total
    print(f"Total recall:\t {(gold_retrieved_passages + gold_retrieved_images) / (total_gold_passages + total_gold_images)}")
    print(f"Total precision:\t {(gold_retrieved_passages + gold_retrieved_images) / (total_retrieved_passages + total_retrieved_images)}")
    print(f"Total f1:\t {2 * (gold_retrieved_passages + gold_retrieved_images) / (total_gold_passages + total_gold_images + total_retrieved_passages + total_retrieved_images)}")



def main(args):
    # Load dataset
    start_time = time.time()
    dataset = load_data_split(args.dataset, args.dataset_split)
    generate_eids = list(range(len(dataset)))
    retriever = Retriever()
    work_eval(args,generate_eids,dataset, retriever)


if __name__ == '__main__':
    if platform.system() == "Darwin":
        multiprocessing.set_start_method('spawn')

    parser = argparse.ArgumentParser()

    # File path or name
    parser.add_argument('--dataset', type=str, default='mmqa',
                        choices=['mmqa'])
    parser.add_argument('--dataset_split', type=str,
                        default='validation', choices=['train', 'validation', 'test'])
    parser.add_argument('--retriever', type=str,default='bd',choices=['dpmlb', 'bd'])

    # debug options
    parser.add_argument('-v', '--verbose', action='store_true', default=False)
    args = parser.parse_args()
    print("Args info:")
    for k in args.__dict__:
        print(k + ": " + str(args.__dict__[k]))
    main(args)



from .model import GPT, turbo, text_davinci_003
from .utils import Logger, fix_seed, print_time, ICLArguments
from .dataset import GPTDataSet, MultiHiertt, FinQA, UnifiedSKG, Wikisql, Wikitq, TAT_QA
from .infer import ICL, turboICL, davinciICL



from typing import Union, Dict
import openai
import time
from abc import ABC, abstractmethod


class GPT(ABC):
    def __init__(self, key: str):
        self.key = key
        openai.api_key = self.key
        self.prompt = None

    def set_prompt(self, prompt: Union[str, Dict[str, str]]):
        self.prompt = prompt

    @abstractmethod
    def getResponse(self, content: str, temperature: float, max_length: int, api_time_interval: float) -> str:
        """
        Get response from the OpenAI model.

        :param content: The input text.
        :param temperature: The temperature for the OpenAI model.
        :param max_length: The maximum length for the OpenAI model response.
        :param api_time_interval: The sleep time for the API (1.0 for text_davinci_003).
        :return: The response text from the OpenAI model.
        """
        pass


class turbo(GPT):
    def addText(self, content: str):
        self.prompt.append({"role": "user", "content": content})

    def getResponse(self, content: str, temperature: float, max_length: int, api_time_interval: float) -> str:
        time.sleep(api_time_interval)
        self.addText(content)
        # response = openai.ChatCompletion.create(
        #     model='gpt-3.5-turbo',
        #     messages=self.prompt,
        #     temperature=temperature
        # )
        # resText = response.choices[0].message.content
        retries = 0
        max_retries = 10
        while retries < max_retries:
            try:
                response = openai.ChatCompletion.create(
                    model='gpt-3.5-turbo',
                    messages=self.prompt,
                    temperature=temperature
                )
                resText = response.choices[0].message.content
                self.deleteText()
                return resText
            except Exception as e:
                print(f"Error: {e}")
                retries += 1
                time.sleep(api_time_interval)
        raise Exception("API call failed after maximum retries.")

    def deleteText(self):
        self.prompt.pop()


class text_davinci_003(GPT):
    def getResponse(self, content: str, temperature: float, max_length: int, api_time_interval: float) -> str:
        time.sleep(api_time_interval)
        # response = openai.Completion.create(
        #     engine="text-davinci-003",
        #     prompt=self.prompt + content,
        #     max_tokens=max_length,
        #     temperature=temperature,
        #     top_p=1,
        #     frequency_penalty=0,
        #     presence_penalty=0,
        #     stop=None
        # )
        # return response["choices"][0]["text"]
        retries = 0
        max_retries = 10
        while retries < max_retries:
            try:
                response = openai.Completion.create(
                    engine="text-davinci-003",
                    prompt=self.prompt + content,
                    max_tokens=max_length,
                    temperature=temperature,
                    top_p=1,
                    frequency_penalty=0,
                    presence_penalty=0,
                    stop=None
                )
                return response["choices"][0]["text"]
            except Exception as e:
                print(f"Error: {e}")
                retries += 1
                time.sleep(api_time_interval)
        raise Exception("API call failed after maximum retries.")



from collections import defaultdict
import re
import string
import json
from tqdm import tqdm
import numpy as np
import pandas as pd
from typing import List, Dict, Tuple
import torch
import random
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import linear_kernel
from .file_utils import is_scatter_available
from tatqa_utils import  *
from .data_util import *
from .data_util import  _is_average, _is_change_ratio, _is_diff, _is_division, _is_sum, _is_times
from nltk.tokenize import sent_tokenize, word_tokenize
import re
import dgl
from .pre_order import infix_to_prefix
from reg_hnt.reghnt.Vocabulary import *
from collections import defaultdict, OrderedDict
from itertools import groupby
from nltk.corpus import stopwords
stop_words = stopwords.words('english')




def get_order_by_tf_idf(question, paragraphs):
    sorted_order = []
    corpus = [question]
    for order, text in paragraphs.items():
        corpus.append(text)
        sorted_order.append(order)
    tf_idf = TfidfVectorizer().fit_transform(corpus)
    cosine_similarities = linear_kernel(tf_idf[0:1], tf_idf).flatten()[1:]
    sorted_similarities = sorted(enumerate(cosine_similarities), key=lambda x:x[1])
    idx = [i[0] for i in sorted_similarities][::-1]
    return [sorted_order[index] for index in idx]



def is_whitespace(c):
    if c == " " or c == "\t" or c == "\r" or c == "\n" or ord(c) == 0x202F:
        return True
    return False

def is_year(item):
    years = [str(i) for i in range(2000, 2051)]
    for year in years:
        if year in item:
            return True
    return False

def is_month(item):
    months = ['January','February','August','March',\
        'April','May','June','July','September',\
            'October','November','December']
    for month in months:
        if month in item:
            return True
    return False

def is_scale(item):
    item = sep_lower(item)
    scales = ['4rxr7y99', 'thousand', 'million', 'billion', 'percent', '%', '£m','USDm','EURm','€m','$m','$M','US$’000','$’000',"£’000",'£000','US$000','$000',"$'000"]
    for scale in scales:
        if scale in item:
            return True
    return False    
    
def is_None(item):
    if item == '' or item == 'N/A' or item == 'n/a' or item == '—' or item == '–' or item == '~'\
    or item == '$—' or item == '—%'or item =='$ —' or item == '— %' or item == "-" or item=='(%)'\
    or item == "$-"or item == "$ -" or item=="-%" or item=="- %"  or item == 'nm' or item=='nm ' or item=='n.m.'\
    or item=="*" or item=="$'000" or item=="$" or item=="%" or item == 'n/m' or item =='$’000'\
    or item=='Years' or item=='£m' or item=='USDm' or item=="EURm" or item == '€m' or item == '$m' or item=='$M'\
    or item=='US$’000' or item=="£’000" or item=='£000' or item=='$000' or item=='US$000' or item=='---':
        return True
    else:
        return False

# answer_dict = {"answer_type": answer_type, "answer": answer, "scale": scale, "answer_from": answer_from}
def generate_class(answer_dict, mapping):
    operator_class = -1
    if answer_dict['answer_type'] == "span":
        if "table" in mapping:
            operator_class = OPERATOR_CLASSES["SPAN-TABLE"]
        else:
            operator_class = OPERATOR_CLASSES["SPAN-TEXT"]
    elif answer_dict['answer_type'] == "multi-span":
        operator_class = OPERATOR_CLASSES["MULTI_SPAN"]
    elif answer_dict['answer_type'] == "count":
        operator_class = OPERATOR_CLASSES["COUNT"]
    elif answer_dict['answer_type'] == "arithmetic":
        operator_class = OPERATOR_CLASSES["ARITHMETIC"]
    return operator_class

def string_tokenizer(string: str, tokenizer) -> List[int]:
    if not string:
        return []
    tokens = []
    string_subword_lens = []
    string_types = []
    prev_is_whitespace = True
    for i, c in enumerate(string):
        if is_whitespace(c):  # or c in ["-", "–", "~"]:
            prev_is_whitespace = True
        else:
            if prev_is_whitespace:
                tokens.append(c)
            else:
                tokens[-1] += c
            prev_is_whitespace = False  # char_to_word_offset.append(len(tokens) - 1)
    split_tokens = []
    for i, token in enumerate(tokens):
        if i != 0:
            sub_tokens = tokenizer.tokenize(" " + token)
        else:
            sub_tokens = tokenizer.tokenize(token)

        for sub_token in sub_tokens:
            split_tokens.append(sub_token)
        string_subword_lens += [len(sub_tokens)]
    ids = tokenizer.convert_tokens_to_ids(split_tokens)
    return ids, string_subword_lens, tokens
'''
scale_type = ['', 'thousand', 'million', 'billion', 'percent'或'%']
'''

'''
table_cell_type_ids  0 none   1 date    2  numbers   3  row_head  4 column_head  5 danwei  -1不参与图网络的，参与训练
table_is_scale_type    0 none    1 scale
'''

def question_tokenize(question, tokenizer):
    if not question:
        return []
    tokens = []
    
    question_word_index = []
    question_subword_lens = []
    question_types = [0]
    question_word_types = []
    current_index = 1
    prev_is_whitespace = True

    question_number_value = []

    question = question.replace("?", '')
    for i, c in enumerate(question):
        if is_whitespace(c):  # or c in ["-", "–", "~"]:
            prev_is_whitespace = True
        else:
            if prev_is_whitespace:
                tokens.append(c)
            else:
                tokens[-1] += c
            prev_is_whitespace = False  # char_to_word_offset.append(len(tokens) - 1)
    split_tokens = []
    for i, token in enumerate(tokens):
        if i != 0:
            sub_tokens = tokenizer.tokenize(" " + token)
        else:
            sub_tokens = tokenizer.tokenize(token)

        flag = 1
        subword_length = 0
        for sub_token in sub_tokens:
            split_tokens.append(sub_token)
            question_word_index.append(current_index)
            flag = 0
            subword_length += 1
            
        question_subword_lens += [subword_length]
        current_index += 1

        if is_number(token) and not is_year(token) and not is_month(tokens[i-1]):  
            question_number_value.append(float(to_number(token)))
            question_types += [2 for _ in range(len(sub_tokens))]
            question_word_types += [2]
        else:
            question_number_value.append(np.nan)
            question_types += [3 for _ in range(len(sub_tokens))]
            question_word_types += [3]

    ids = tokenizer.convert_tokens_to_ids(split_tokens)
    question_types += [0]    
    return ids, tokens, question_subword_lens, question_word_index, question_types, question_word_types,question_number_value


def clean_year(item):
    flag = 0
    years = [str(i) for i in range(2000, 2051)]
    num = 0
    this_year = ''
    for year in years:
        if year in item:
            num += 1
            this_year = year
    if num != 1 or item == '2015/16' or item == '2016/17' or item == '2017/18' or item == '2018/19' or item == '2019/20' or item =="2025-beyond":
        return item
    else:
        if is_month(item):
            return item
        else:
            if this_year != item:
                import pdb; pdb.set_trace()
            return this_year        

def scale_judge(item):
    item = sep_lower(item)
    scales = ['4rxr7y99', 'thousand', 'million', 'billion', 'percent', '%', '£m','USDm','EURm','€m','$m','$M','US$’000','$’000',"£’000",'£000','US$000','$000',"$'000"]
    for idx, scale in enumerate(scales):
        if scale in item:
            to_idx = 0
            if idx in [1, 12, 13, 14, 15, 16, 17, 18]:
                to_idx = 1
            elif idx in [2, 6, 7, 8, 9, 10, 11]:
                to_idx = 2
            elif idx in [3]:
                to_idx = 3
            elif idx in [4, 5]:
                to_idx = 4
            return idx, to_idx

def generate_number_scale(table, table_pd, table_cell_type_ids, table_is_scale_type, table_cell_number_value, table_cell_number_value_position, paragraphs):
    SCALE = ['', 'thousand', 'million', 'billion', 'percent']
    TMB_SCALE = ['thousand', 'million', 'billion']

    table_number_position = []
    for x, y in zip(table_cell_number_value_position, table_cell_number_value):
        if not np.isnan(y):
            table_number_position.append(x)
        else:
            table_number_position.append((-1, -1))
        
        
    table_cell_number_scale = list(map(lambda x: 0 if not np.isnan(x) else x, table_cell_number_value))
    global_scale = 0

    scale_position = np.where(table_is_scale_type > 0)
    for x, y in zip(*scale_position):
        
        this_scale, to_scale = scale_judge(table[x][y])
        if this_scale in [1, 2, 3, 6, 7]:
            global_scale = to_scale
            
            for idx in range(x, len(table)):
                for idy in range(len(table[idx])):
                    if (idx, idy) in table_cell_number_value_position:
                        table_cell_number_scale[table_cell_number_value_position.index((idx, idy))] = to_scale
        

    if global_scale == 0:
        for paragraph in paragraphs:
            paragraph_text = paragraph['text']
            if 'in thousand' in paragraph_text or 'In thousands' in paragraph_text:
                global_scale = 1
            if 'in million' in paragraph_text or 'In million' in paragraph_text:
                global_scale = 2
            if 'in billion' in paragraph_text or 'In billion' in paragraph_text:
                global_scale = 3
        if global_scale != 0:
            table_cell_number_scale = list(map(lambda x:global_scale if not np.isnan(x) else x, table_cell_number_scale))
            
    for x, y in zip(*scale_position):
        this_scale, to_scale = scale_judge(table[x][y])
        if this_scale in [4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17]:
            for item in table_number_position:
                if item[1] == y and item[0] > x:
                    table_cell_number_scale[table_cell_number_value_position.index(item)] = to_scale
                if item[0] == x and item[1] > y:
                    table_cell_number_scale[table_cell_number_value_position.index(item)] = to_scale
        
        
    for i in range(len(table)):
        for j in range(len(table[i])):
            if table_cell_type_ids[i][j] == 2:
                if '%' in table[i][j]:
                    table_cell_number_scale[table_cell_number_value_position.index((i, j))] = 4
    
    return table_cell_number_scale
            
        
            
            
    


def table_tokenize(table, tokenizer, question_id):
    table_pd = pd.DataFrame(table, dtype=np.str)
    table_cell_type_ids = np.full(table_pd.shape, -1)
    table_is_scale_type = np.zeros(table_pd.shape)
    table_cell_tokens = []
    table_cell_number_value = []
    table_cell_number_value_position = []
    table_cell_number_value_position_b = []
    table_ids = []
    table_cell_index = []
    table_subword_lens = []
    table_types = []
    table_word_types = []
    current_cell_index = 0
    table_word_tokens = []
    table_word_tokens_number = []
    table_cell_types_flatten = []
    
    row_dict = defaultdict(list)
    column_dict = defaultdict(list)
    
    for i in range(len(table)):
        for j in range(len(table[i])):
            
            #一些特殊的处理
            table[i][j] = table[i][j].replace('bps', '')

            if table[i][j] in ['$—', '—%', '$ —', '— %', "-","—","$-","$ -","-%","- %"]:
                table[i][j] = '0'
            
            if is_None(table[i][j]):
                table_cell_type_ids[i][j] = 0
            elif is_number(table[i][j]) and not is_year(table[i][j]):
                table_cell_type_ids[i][j] = 2
            elif is_year(table[i][j]):
                table_cell_type_ids[i][j] = 1
            else:
                flag1 = 0
                for item in table[i]:
                    if item != '':
                        flag1 += 1
                
                flag = 0
                for item in table[i]:
                    if is_number(item) and not is_year(item):
                        flag = 1
                if flag:
                    table_cell_type_ids[i][j] = 3
                flag = 0
                for item in [item[j] for item in table]:
                    if is_number(item) and not is_year(item):
                        flag = 1
                if flag and flag1 != 1:
                    table_cell_type_ids[i][j] = 4
                
            if is_scale(table[i][j]) and not is_number(table[i][j]):
                table_is_scale_type[i][j] = 1
            
            if table_cell_type_ids[i][j] == 0:
                continue  
            table_cell_input_token = table[i][j]
            
                
            table_cell_tokens.append(table[i][j])
            
            if table_cell_type_ids[i][j] == 2:
                per_flag = 0
                for c_idx in range(len([item[j] for item in table])):
                    if table_is_scale_type[c_idx][j] == 1 and scale_judge(table[c_idx][j])==4:
                        per_flag = 1
                for c_idx in range(len(table[i])):
                    if table_is_scale_type[i][c_idx] == 1 and scale_judge(table[i][c_idx])==4:
                        per_flag = 1
                num_number = to_number(table[i][j]) 
                if per_flag and '%' not in table[i][j]:
                    num_number  = round(float(num_number/100.0),4)
                table_cell_number_value.append(num_number)
                table_cell_input_token = table[i][j].replace(' ','').replace('$','')
                if '(' in table_cell_input_token and ')' in table_cell_input_token:
                    table_cell_input_token = table_cell_input_token.replace('(', '').replace(')', '')
                    table_cell_input_token = '-' + table_cell_input_token
                table_cell_number_value_position.append((i, j))
            else:
                table_cell_number_value.append(np.nan)
                table_cell_number_value_position.append((-1, -1))
            cell_ids, string_subword_lens, tokens= string_tokenizer(table_cell_input_token, tokenizer)
            table_word_tokens.extend(tokens)
            table_word_tokens_number += [len(tokens)]
            table_subword_lens += string_subword_lens
            
            if table_cell_type_ids[i][j] == 2:
                table_types += [2 for _ in range(len(cell_ids))]
                table_word_types += [2 for _ in range(len(string_subword_lens))]
            else:
                table_types += [3 for _ in range(len(cell_ids))]
                table_word_types += [3 for _ in range(len(string_subword_lens))]
                
            table_ids += cell_ids
            table_cell_index += [current_cell_index for _ in range(len(cell_ids))]

            
            row_dict[i].append(current_cell_index)
            column_dict[j].append(current_cell_index)
            table_cell_types_flatten.append(table_cell_type_ids[i][j])
            current_cell_index += 1

    return table_pd,table_cell_type_ids,table_is_scale_type,table_cell_tokens,table_cell_number_value,\
        table_ids,table_cell_index, table_subword_lens, table_types, table_word_types,table_word_tokens,table_word_tokens_number,table_cell_number_value_position,\
            row_dict, column_dict, table_cell_types_flatten


def paragraph_number_judge(tokens, idx, this_number, paragraphs):
    global_scale = 0
    
    for i, paragraph_text in paragraphs.items():
        if 'in thousand' in paragraph_text:
            global_scale = 1
        if 'in million' in paragraph_text:
            global_scale = 2
        if 'in billion' in paragraph_text:
            global_scale = 3

    
    flag = 1
    scale_flag = 0
    if idx != 0:
        if is_month(tokens[idx-1]) and '$' not in this_number:
            flag = 0
    if idx != len(tokens) - 1:
        if is_month(tokens[idx+1]) and '$' not in this_number:
            flag = 0 
    if idx == 0:
        flag = 0


    if this_number[-1] == '.' and '$' not in this_number and '%' not in this_number and '£' not in this_number:
        flag = 0
                
    
    for stopword in ['(1)','(2)','(3)','(4)','(5)','(6)','(7)']:
        if stopword in this_number:
            flag = 0
            break
    
    if flag != 0:
        if '%' in this_number:
            scale_flag = 4
            
        if 'm' in this_number:
            scale_flag = 2
        if 'b' in this_number:
            scale_flag = 3
        
        for token in tokens[idx+1:idx+3]:
            token = sep_lower(token)
            for scale_idx, scale_label in enumerate(['dfe526f2', 'thousand', 'million', 'billion', 'percent']):
                if scale_label in token:
                    scale_flag = scale_idx
    
    if scale_flag == 0 and global_scale != 0:
        scale_flag = global_scale
    
    return flag, scale_flag
        
    
def paragraph_tokenize(question, paragraphs, tokenizer, question_id):
    paragraphs_copy = paragraphs.copy()
    paragraphs = {}
    for paragraph in paragraphs_copy:
        paragraphs[paragraph["order"]] = paragraph["text"]
    del paragraphs_copy
    split_tokens = []

    number_mask = []
    number_value = []

    tokens = []

    paragraph_subword_tags = []
    paragraph_index = []
    paragraph_tags = []
    
    paragraph_subword_lens = []
    paragraph_types = []
    paragraph_word_types = []
    paragraph_number_scale = []
    paragraph_sep_tag = []
    paragraph_mapping = False
    paragraph_mapping_orders = []

    sorted_order = get_order_by_tf_idf(question, paragraphs)

    last_token_length = 0
    for order in sorted_order:
        text = paragraphs[order]
        text = " ".join(text.split())
        prev_is_whitespace = True
        answer_indexs = None

        start_index = 0
        wait_add = False

        tokens.append(tokenizer.sep_token)

        for i, c in enumerate(text):
            if is_whitespace(c):
                if wait_add:
                    wait_add = False
                prev_is_whitespace = True
            else:
                if prev_is_whitespace:
                    tokens.append(c)
                    wait_add = True
                    start_index = i
                else:
                    tokens[-1] += c
                prev_is_whitespace = False
    if "direct‐to‐grower" in tokens:
        tokens.remove("direct‐to‐grower")
    current_token_index = 0
    for i, token in enumerate(tokens):
        if token == tokenizer.sep_token:
            sub_tokens = tokenizer.tokenize(token)
            paragraph_sep_tag += [1]
        elif i != 0:
            sub_tokens = tokenizer.tokenize(" " + token)
            paragraph_sep_tag += [0]
        else:
            sub_tokens = tokenizer.tokenize(token)
            paragraph_sep_tag += [0]
        paragraph_subword_lens += [len(sub_tokens)]
        stop_flag = 0
        if is_number(token) and not is_year(token):
            stop_flag, scale_flag = paragraph_number_judge(tokens, i, token, paragraphs)
        if stop_flag == 1 and is_number(token) and not is_year(token):
            number_value.append(float(to_number(token)))
            paragraph_types += [2 for _ in range(len(sub_tokens))]
            paragraph_word_types += [2]
            paragraph_number_scale.append(scale_flag)
        else:
            number_value.append(np.nan)
            paragraph_types += [3 for _ in range(len(sub_tokens))]
            paragraph_word_types += [3]
            paragraph_number_scale.append(np.nan)

        for sub_token in sub_tokens:
            split_tokens.append(sub_token)
            paragraph_index.append(current_token_index)

        current_token_index+=1
        paragraph_subword_tags += [1]
        if len(sub_tokens) > 1:
            paragraph_subword_tags += [0] * (len(sub_tokens) - 1)
    paragraph_ids = tokenizer.convert_tokens_to_ids(split_tokens)

    return tokens, paragraph_ids, number_value, paragraph_index,paragraph_subword_lens,paragraph_types,paragraph_word_types, paragraph_number_scale, paragraph_sep_tag
    
    
def subword_judge(subword_lens, input_types):
    aa = len(input_types) - len(torch.where(torch.tensor(input_types)==0)[0])
    bb = 0
    for i in subword_lens:
        bb += i
    if aa == bb:
        return subword_lens
    else:
        cc = bb - aa
        subword_lens[-1] -= cc
        return subword_lens
    

def _concat(question_ids,
            table_ids, 
            paragraph_ids, 
            question_tokens,
            table_tokens,
            paragraph_tokens,
            table_cell_number_value,
            paragraph_number_value,
            table_cell_index,
            paragraph_index,
            question_subword_lens,
            question_types,
            table_subword_lens,
            table_types,
            paragraph_subword_lens,
            paragraph_types,
            question_word_types,
            table_word_types,
            paragraph_word_types,
            table_cell_number_value_position,
            table_cell_number_scale,
            paragraph_number_scale,
            question_number_value,
            row_dict,
            column_dict,
            table_cell_types_flatten,
            sep,
            question_length_limitation,
            passage_length_limitation,
            max_pieces):
    
    input_ids = torch.zeros([1,max_pieces])
    token_type_ids = torch.zeros_like(input_ids)

    question_ids = [sep] + question_ids + [sep]
    question_length = len(question_ids)
    table_length = len(table_ids)
    paragraph_length = len(paragraph_ids)
    table_cell_count_origin = len(table_cell_types_flatten)
    if passage_length_limitation is not None:
        passage_length_limitation = passage_length_limitation - question_length - 2
        if len(table_ids) > passage_length_limitation:
            passage_ids = table_ids[:passage_length_limitation]
            table_length = passage_length_limitation
            paragraph_length = 0
            table_cell_index = table_cell_index[:passage_length_limitation]
            table_types = table_types[:passage_length_limitation]
            end = table_cell_index[-1]
            table_subword_lens = table_subword_lens[:end+1]
            table_word_types = table_word_types[:end+1]
            table_tokens = table_tokens[:end+1]
            table_cell_types_flatten = table_cell_types_flatten[:end+1]
            table_cell_number_value = table_cell_number_value[:end+1]
            table_cell_number_value_position = table_cell_number_value_position[:end+1]
            table_cell_number_scale = table_cell_number_scale[:end+1]
        elif len(table_ids) + len(paragraph_ids) > passage_length_limitation:
            passage_ids = table_ids + [sep] + paragraph_ids
            passage_ids = passage_ids[:passage_length_limitation]
            paragraph_index=paragraph_index[:passage_length_limitation-table_length-1]
            paragraph_types = paragraph_types[:passage_length_limitation-table_length-1]
            end = paragraph_index[-1]
            paragraph_tokens = paragraph_tokens[:end+1]
            paragraph_number_value = paragraph_number_value[:end+1]
            paragraph_number_scale = paragraph_number_scale[:end+1]
            paragraph_subword_lens = paragraph_subword_lens[:end+1]
            paragraph_word_types = paragraph_word_types[:end+1]
            table_length = len(table_ids)
            paragraph_length = passage_length_limitation - table_length
        else:
            passage_ids = table_ids + [sep] + paragraph_ids
            table_length = len(table_ids)
            paragraph_length = len(paragraph_ids) + 1
    else:
        passage_ids = table_ids + [sep] + paragraph_ids
    
    passage_ids = passage_ids + [sep]
    input_ids[0, :question_length] = torch.from_numpy(np.array(question_ids))
    input_ids[0, question_length:question_length + len(passage_ids)] = torch.from_numpy(np.array(passage_ids))
    attention_mask = input_ids != 0

    input_types = question_types + table_types
    if paragraph_length > 1:
        input_types += [0] + paragraph_types
    
    
    try:
        table_number = [item for item in table_cell_number_value if not np.isnan(item)]
    except:
        import pdb; pdb.set_trace()
    
    paragraph_number = [item for item in paragraph_number_value if not np.isnan(item)]
    question_number = [item for item in question_number_value if not np.isnan(item)]
    table_number_position = [x for x, y in zip(table_cell_number_value_position, table_cell_number_value) if not np.isnan(y)]
    
    table_number_index = [i for i,item in enumerate(table_cell_number_value) if not np.isnan(item)]
    paragraph_number_index = [i for i,item in enumerate(paragraph_number_value) if not np.isnan(item)]
    question_number_index = [i for i,item in enumerate(question_number_value) if not np.isnan(item)]

    table_number_scale = [item for i,item in enumerate(table_cell_number_scale) if not np.isnan(item)]
    paragraph_number_scale = [item for i,item in enumerate(paragraph_number_scale) if not np.isnan(item)]
    
    subword_lens = question_subword_lens + table_subword_lens + paragraph_subword_lens
    input_word_types = question_word_types + table_word_types + paragraph_word_types
    
    
    subword_lens = subword_judge(subword_lens, input_types)
    
    table_cell_count = len(table_cell_types_flatten)
    if table_cell_count_origin != table_cell_count:
        import pdb; pdb.set_trace()
        
    for key, key_list in row_dict.items():
        for item in key_list:
            if item >= table_cell_count:
                key_list.remove(item)
    for key, key_list in column_dict.items():
        for item in key_list:
            if item >= table_cell_count:
                key_list.remove(item)
    
    
    return input_ids, attention_mask, input_types, input_word_types, subword_lens,\
        table_cell_index,table_tokens,table_number, table_number_index, paragraph_index,\
            paragraph_tokens, paragraph_number, paragraph_number_index,token_type_ids,table_number_position,\
                table_number_scale, paragraph_number_scale, question_number, question_number_index, row_dict, column_dict, table_cell_types_flatten



OPERATOR = ['+', '-', '*', '/']

def get_operators(derivation:str):
    res = []
    for c in derivation:
        if c in OPERATOR:
            res.append(c)
    return res

def delete_punc(example):
    if example == '[SEP]' or example == '</s>' or example == '<s>':
        return example
    punctuation_string = string.punctuation
    for i in punctuation_string:
        example = example.replace(i, '')
    return example
def sep_lower(example):
    if example == '[SEP]' or example == '</s>' or example == '<s>':
        return example
    else:
        return example.lower()


class GraphFactory(object):
    def __init__(self, question_tokens, table_tokens, paragraph_tokens, row_dict, column_dict, table_types, table, table_origin_types, tokenizer
        ):
        self.question_tokens = question_tokens
        self.table_tokens = table_tokens
        self.paragraph_tokens = paragraph_tokens
        self.global_tokens = question_tokens + table_tokens + paragraph_tokens
        self.table_types = table_types
        self.row_dict = row_dict
        self.column_dict = column_dict
        self.row_dict = OrderedDict(sorted(self.row_dict.items(), key=lambda x: x[0], reverse=False))
        self.column_dict = OrderedDict(sorted(self.column_dict.items(), key=lambda x: x[0], reverse=False))
        self.graph = defaultdict(list)
        self.q_n = len(question_tokens)
        self.t_n = len(table_tokens)
        self.p_n = len(paragraph_tokens)
        self.table = table
        self.table_origin_types = table_origin_types
        self.tokenizer = tokenizer
    
    def generate_graph(self):
        inner_table_relations = self.generate_inner_table_relation()
        inner_paragraph_relations = self.generate_inner_paragraph_relation()
        inner_question_relations = self.generate_inner_question_relation()
        question_table_relations = self.generate_question_table_relation()
        question_paragraph_relations = self.generate_question_paragraph_relation()
        paragraph_table_relations = self.generate_paragraph_table_relation()
        global_relations = inner_table_relations+inner_paragraph_relations+\
            inner_question_relations+question_table_relations+question_paragraph_relations+\
                paragraph_table_relations
        global_relations = sorted(global_relations, key=lambda x:(x[0],x[1]))
        
        global_relations_new = []
        punctuation_string = string.punctuation
        for relation in global_relations:
            string1 = self.global_tokens[relation[0]]
            string2 = self.global_tokens[relation[1]]
            
            if string1 in punctuation_string or string2 in punctuation_string:
                if not ((relation[0] > self.q_n+self.t_n \
                and relation[1] > self.q_n+self.t_n)
                or (relation[0] <= self.q_n \
                and relation[1] <= self.q_n)\
                or (relation[0] <= self.q_n+self.t_n and relation[0]>self.q_n  \
                and relation[1] <= self.q_n+self.t_n and relation[0]>self.q_n )
                    ):
                    continue
        
            if relation[0]>self.q_n+self.t_n+self.p_n or \
                relation[1]>self.q_n+self.t_n+self.p_n:
                continue
            global_relations_new.append(relation)
        self.relations = sorted(global_relations_new, key=lambda x:(x[0],x[1]))
        
        
        target_set = set(list(range(0, self.q_n+self.t_n+self.p_n)))
        current_set = set()
        for relation in self.relations:
            current_set.add(relation[0])
        cha = list(target_set-current_set)
        if cha:
            for chas in cha:
                self.relations.append((chas, chas, "self-same-self"))
        relation_set = list(set(self.relations))
        if len(relation_set) != len(self.relations):
            import pdb; pdb.set_trace()        
        
        graph = {}
        graph['src'] = list(map(lambda x:x[0], self.relations))
        graph['dst'] = list(map(lambda x:x[1], self.relations))
        graph['relations'] = list(map(lambda x:Relation_vocabulary['word2id'][x[2]], self.relations))
        graph['triplet'] = self.relations
        graph['dgl'] = dgl.graph((graph['src'], graph['dst']))
    
        return graph
        
        
    def type_check(self, idx):
        return self.table_types[idx]
    
    # 日期 1， row_head 3  column_head 4, number 2
    def generate_inner_table_relation(self):
        self.inner_table_relation = []
        
        for Row, R_list in self.row_dict.items():
            for i in range(len(R_list)):
                for j in range(len(R_list)):
                    if i == j:
                        continue
                    if self.type_check(R_list[i]) == 3 and self.type_check(R_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n+R_list[i], self.q_n+R_list[j], "Row-Contain-Cell"))
                    elif self.type_check(R_list[i]) == 2 and self.type_check(R_list[j]) == 3:
                        self.inner_table_relation.append((self.q_n+R_list[i], self.q_n+R_list[j], "Cell-Belong-Row"))
                    elif self.type_check(R_list[i]) == 1 and self.type_check(R_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n+R_list[i], self.q_n+R_list[j], "Time-Contain-Cell"))
                    elif self.type_check(R_list[i]) == 2 and self.type_check(R_list[j]) == 1:
                        self.inner_table_relation.append((self.q_n+R_list[i], self.q_n+R_list[j], "Cell-Belong-Time"))
                    elif self.type_check(R_list[i]) == 2 and self.type_check(R_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n+R_list[i], self.q_n+R_list[j], "Cell-SameRow-Cell"))
        for Column, C_list in self.column_dict.items():
            for i in range(len(C_list)):
                for j in range(len(C_list)):
                    if i == j:
                        continue
                    if self.type_check(C_list[i]) == 4 and self.type_check(C_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n+C_list[i], self.q_n+C_list[j], "Column-Contain-Cell"))
                    elif self.type_check(C_list[i]) == 2 and self.type_check(C_list[j]) == 4:
                        self.inner_table_relation.append((self.q_n+C_list[i], self.q_n+C_list[j], "Cell-Belong-Column"))
                    elif self.type_check(C_list[i]) == 1 and self.type_check(C_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n+C_list[i], self.q_n+C_list[j], "Time-Contain-Cell"))
                    elif self.type_check(C_list[i]) == 2 and self.type_check(C_list[j]) == 1:
                        self.inner_table_relation.append((self.q_n+C_list[i], self.q_n+C_list[j], "Cell-Belong-Time"))

        xx, yy = self.table_origin_types.shape
        special_year = 0
        for i, row in enumerate(self.table_origin_types):
            if row[0] == 1 and np.sum(row) == 1 and i != 0:
                special_year += 1
        if special_year > 1:
            table_types_numpy = np.array(self.table_types)
            year_tags = np.where(table_types_numpy==1)[0]
            for i in range(len(year_tags)-1):
                start = year_tags[i]
                for j in range(year_tags[i]+1, year_tags[i+1]):
                    if self.table_types[j] == 2:
                        if (self.q_n+start, self.q_n+j, "Time-Contain-Cell") not in self.inner_table_relation:
                            self.inner_table_relation.append((self.q_n+start, self.q_n+j, "Time-Contain-Cell"))
                        if (self.q_n+j, self.q_n+start, "Cell-Belong-Time") not in self.inner_table_relation:
                            self.inner_table_relation.append((self.q_n+j, self.q_n+start, "Cell-Belong-Time"))
            start = year_tags[-1]
            for j in range(year_tags[-1]+1, len(self.table_types)):
                if self.table_types[j] == 2:
                    if (self.q_n+start, self.q_n+j, "Time-Contain-Cell") not in self.inner_table_relation:
                        self.inner_table_relation.append((self.q_n+start, self.q_n+j, "Time-Contain-Cell"))
                    if (self.q_n+j, self.q_n+start, "Cell-Belong-Time") not in self.inner_table_relation:
                        self.inner_table_relation.append((self.q_n+j, self.q_n+start, "Cell-Belong-Time"))

        self.inner_table_relation = sorted(self.inner_table_relation, key=lambda x:(x[0],x[1]))

        return self.inner_table_relation
    
    def generate_inner_paragraph_relation(self):
        self.inner_paragraph_relation = []
        select_node = []
        for idx, node in enumerate(self.paragraph_tokens):
            if node == self.tokenizer.sep_token:
                select_node.append(idx)
        paragraph_split = []
        last_select_node = 0
        for idx, node in enumerate(self.paragraph_tokens):
            if node == self.tokenizer.sep_token:
                last_select_node = idx
            if node != self.tokenizer.sep_token:
                self.inner_paragraph_relation.append((self.q_n+self.t_n+last_select_node, self.q_n+self.t_n+idx, "Sentence-Contain-P_Word"))
                self.inner_paragraph_relation.append((self.q_n+self.t_n+idx, self.q_n+self.t_n+last_select_node, "P_Word-Partof-Sentence"))
        for i in range(len(self.paragraph_tokens) - 1):
            if self.paragraph_tokens[i] != self.tokenizer.sep_token and self.paragraph_tokens[i+1] != self.tokenizer.sep_token:
                self.inner_paragraph_relation.append((self.q_n+self.t_n+i, self.q_n+self.t_n+i+1, "Q_Word-PriorWord-Q_Word"))
                self.inner_paragraph_relation.append((self.q_n+self.t_n+i+1, self.q_n+self.t_n+i, "Q_Word-NextWord-Q_Word"))
        self.inner_paragraph_relation = sorted(self.inner_paragraph_relation, key=lambda x:(x[0],x[1]))
        return self.inner_paragraph_relation
    
    def generate_inner_question_relation(self):
        self.inner_question_relations = []
        for i in range(len(self.question_tokens) - 1):
            self.inner_question_relations.append((i, i+1, "Q_Word-PriorWord-Q_Word"))
            self.inner_question_relations.append((i+1, i, "Q_Word-NextWord-Q_Word"))
        self.inner_question_relations = sorted(self.inner_question_relations, key=lambda x:(x[0],x[1]))
        return self.inner_question_relations

    def generate_question_table_relation(self):
        question = " ".join(self.question_tokens)
        question = question.lower()
        self.question_table_relations = []
        
        question_tokens = list(map(delete_punc, self.question_tokens))
        question_tokens = list(map(sep_lower, question_tokens))
        table_tokens = list(map(delete_punc, self.table_tokens))
        table_tokens = list(map(sep_lower, table_tokens))
        
        for i, word in enumerate(question_tokens):
            word = sep_lower(word)
            if word in stop_words:
                continue
            for j, cell in enumerate(table_tokens):
                flag1, flag2 = 0, 0
                cell = cell.strip(':')
                cell = sep_lower(cell)
                cell_words = cell.split(' ')
                if self.table_types[j] == 2:
                    continue
                if word in cell_words:
                    flag1 = 1
                if cell in question:
                    flag2 = 1
                if flag1 and flag2:
                    self.question_table_relations.append((i, self.q_n+j, "Q_Word-ExactMatch-Row"))
                    self.question_table_relations.append((self.q_n+j, i, "Row-ExactMatch-Q_Word"))
                elif flag1 and not flag2:
                    self.question_table_relations.append((i, self.q_n+j, "Q_Word-PartialMatch-Row"))
                    self.question_table_relations.append((j+self.q_n, i, "Row-PartialMatch-Q_Word"))
        self.question_table_relations = sorted(self.question_table_relations, key=lambda x:(x[0],x[1]))

        return self.question_table_relations
                
    def generate_question_paragraph_relation(self):
        self.question_paragraph_relations = []
        question_tokens = list(map(delete_punc, self.question_tokens))
        question_tokens = list(map(sep_lower, question_tokens))
        paragraph_tokens = list(map(delete_punc, self.paragraph_tokens))
        paragraph_tokens = list(map(sep_lower, paragraph_tokens))
        
        select_node = []
        for idx, node in enumerate(paragraph_tokens):
            if node == self.tokenizer.sep_token:
                select_node.append(idx)
        split_number = len(select_node)
        
        paragraph_split = [list(g) for k, g in groupby(paragraph_tokens, lambda x:x==self.tokenizer.sep_token) if not k]

        
        for i, paragraph in enumerate(paragraph_split):
            for j, word in enumerate(question_tokens):
                if word in stop_words or word == '':
                    continue
                if word in paragraph:
                    self.question_paragraph_relations.append((j, select_node[i]+self.q_n+self.t_n, "Q_Word-Partof-Sentence"))
                    self.question_paragraph_relations.append((select_node[i]+self.q_n+self.t_n, j, "Sentence-Contain-Q_Word"))
        
        for i, p_word in enumerate(paragraph_tokens):
            if p_word in stop_words or p_word == '':
                continue
            for j, word in enumerate(question_tokens):
                if p_word == word:
                    self.question_paragraph_relations.append((j, i+self.q_n+self.t_n, "Word-Same-Word"))
                    self.question_paragraph_relations.append((i+self.q_n+self.t_n, j, "Word-Same-Word"))
        self.question_paragraph_relations = sorted(self.question_paragraph_relations, key=lambda x:(x[0],x[1]))
        return self.question_paragraph_relations
    
    def generate_paragraph_table_relation(self):
        self.paragraph_table_relations = []
        table_tokens = list(map(delete_punc, self.table_tokens))
        table_tokens = list(map(sep_lower, table_tokens))
        paragraph_tokens = list(map(delete_punc, self.paragraph_tokens))
        paragraph_tokens = list(map(sep_lower, paragraph_tokens))
        
        select_node = []
        for idx, node in enumerate(paragraph_tokens):
            if node == self.tokenizer.sep_token:
                select_node.append(idx)
        split_number = len(select_node)
        
        paragraph_split = [list(g) for k, g in groupby(paragraph_tokens, lambda x:x==self.tokenizer.sep_token) if not k]
        
        for id, paragraph_tokens_new in enumerate(paragraph_split):
            paragraph_tokens_new_new = paragraph_tokens_new.copy()
            a = paragraph_tokens_new_new.count('')
            for _ in range(a):
                paragraph_tokens_new_new.remove('')
            paragraph = " ".join(paragraph_tokens_new_new)
            for i, word in enumerate(paragraph_tokens_new):
                word = sep_lower(word)
                if word in stop_words:
                    continue
                for j, cell in enumerate(table_tokens):
                    flag1, flag2 = 0, 0
                    cell = cell.strip(':')
                    cell = sep_lower(cell)
                    cell_words = cell.split(' ')
                    if self.table_types[j] == 2:
                        continue
                    if word in cell_words:
                        flag1 = 1
                    if cell in paragraph:
                        flag2 = 1
                    if flag1 and flag2:
                        self.paragraph_table_relations.append((select_node[id]+i+1+self.q_n+self.t_n, j+self.q_n, "P_Word-ExactMatch-Row"))
                        self.paragraph_table_relations.append((j+self.q_n, select_node[id]+i+1+self.q_n+self.t_n, "Row-ExactMatch-P_Word"))
                    elif flag1 and not flag2:
                        self.paragraph_table_relations.append((select_node[id]+i+1+self.q_n+self.t_n, j+self.q_n, "P_Word-PartialMatch-Row"))
                        self.paragraph_table_relations.append((j+self.q_n, select_node[id]+i+1+self.q_n+self.t_n, "Row-PartialMatch-P_Word"))
            for k, cell2 in enumerate(table_tokens):
                cell2 = cell2.strip(':')
                cell2 = sep_lower(cell2)
                cell_words = cell2.split(' ')
                if self.table_types[k] == 2:
                    continue
                if cell2 in paragraph:
                    self.paragraph_table_relations.append((select_node[id]+self.q_n+self.t_n, k+self.q_n, "Sentence-Contain-Row"))
                    self.paragraph_table_relations.append((k+self.q_n, select_node[id]+self.q_n+self.t_n, "Row-Partof-Sentence"))
        self.paragraph_table_relations = sorted(self.paragraph_table_relations, key=lambda x:(x[0],x[1]))             
        return self.paragraph_table_relations
        



class TaTQATestReader(object):
    def __init__(self, tokenizer, 
                 passage_length_limit: int = None, question_length_limit: int = None, sep="<s>"):
        self.max_pieces = 512
        self.tokenizer = tokenizer
        self.passage_length_limit = passage_length_limit
        self.question_length_limit = question_length_limit
        self.sep = self.tokenizer.convert_tokens_to_ids(sep)
        self.skip_count = 0
    
    def _make_instance(self,input_ids, attention_mask,token_type_ids,
                question_tokens,
                table, table_cell_type_ids,table_is_scale_type, table_number, table_number_index, 
                table_cell_index, table_tokens,paragraph_tokens,
                paragraph_index,paragraph_number,paragraph_number_index,
                question_id,
                word_mask, number_mask, subword_lens, item_length,
                word_word_mask, number_word_mask, table_word_tokens, table_word_tokens_number,
                table_number_scale, paragraph_number_scale, graph, question, question_number,question_number_index):
        return {
            "input_ids": np.array(input_ids),
            "attention_mask": np.array(attention_mask),
            "token_type_ids": np.array(token_type_ids),

            "question_tokens": question_tokens,

            "table": table,
            "table_cell_type_ids": table_cell_type_ids,
            "table_is_scale_type": table_is_scale_type,
            "table_number": table_number,
            "table_number_index": np.array(table_number_index),
            "table_cell_index": np.array(table_cell_index),
            "table_tokens": table_tokens,

            "paragraph_tokens": paragraph_tokens,
            "paragraph_index": np.array(paragraph_index),
            "paragraph_number": paragraph_number,
            "paragraph_number_index": np.array(paragraph_number_index),

            "question_id": question_id,
            "word_mask": np.array(word_mask),
            "number_mask": np.array(number_mask),
            "subword_lens": np.array(subword_lens),
            "item_length": item_length,
            
            "word_word_mask": np.array(word_word_mask),
            "number_word_mask": np.array(number_word_mask),
            
            "table_word_tokens": table_word_tokens,
            
            "table_word_tokens_number": table_word_tokens_number,
            
            "table_number_scale": table_number_scale,
            "paragraph_number_scale" :paragraph_number_scale,
            
            "graph": graph,
            
            "question": question,

            "question_number": question_number,
            "question_number_index": question_number_index
        }        
    
    def _to_instance(self, question: str, table: List[List[str]], paragraphs: List[Dict], question_id:str):
        
        table_pd,table_cell_type_ids,table_is_scale_type,table_cell_tokens,\
        table_cell_number_value,table_ids,\
        table_cell_index, table_subword_lens, table_types, table_word_types,\
        table_word_tokens, table_word_tokens_number, table_cell_number_value_position,\
        row_dict, column_dict, table_cell_types_flatten = table_tokenize(table, self.tokenizer,question_id)
        

        
        paragraph_tokens, paragraph_ids, number_value, \
        paragraph_index, paragraph_subword_lens,paragraph_types, paragraph_word_types, paragraph_number_scale, paragraph_sep_tag\
        = paragraph_tokenize(question, paragraphs, self.tokenizer, question_id)
        
        question_ids, question_tokens, question_subword_lens, question_word_index, question_types, question_word_types, question_number_value\
        = question_tokenize(question, self.tokenizer)
        
        table_cell_number_scale = generate_number_scale(table, table_pd, table_cell_type_ids, table_is_scale_type, table_cell_number_value, table_cell_number_value_position, paragraphs)


        input_ids, attention_mask, input_types, input_word_types, subword_lens, \
        table_cell_index,table_tokens,table_number, table_number_index, paragraph_index,\
        paragraph_tokens, paragraph_number, paragraph_number_index,token_type_ids, \
        table_number_position,table_number_scale, paragraph_number_scale,question_number, question_number_index,\
        row_dict, column_dict, table_cell_types_flatten\
        = _concat(question_ids, table_ids, paragraph_ids,\
        question_tokens,table_cell_tokens, paragraph_tokens,table_cell_number_value,number_value,\
        table_cell_index, paragraph_index,question_subword_lens,question_types,table_subword_lens,\
        table_types, paragraph_subword_lens,paragraph_types,question_word_types,table_word_types,\
        paragraph_word_types,table_cell_number_value_position,table_cell_number_scale, paragraph_number_scale, question_number_value, \
        row_dict, column_dict, table_cell_types_flatten,
        self.sep,self.question_length_limit,self.passage_length_limit, self.max_pieces)
        
        
        item_length = {'q': len(question_tokens), 't': len(table_tokens), 'p':len(paragraph_tokens)}
        
        word_mask = torch.zeros(512)
        number_mask = torch.zeros(512)
        
        try:
            for i, idx in enumerate(input_types):
                if idx == 3:
                    word_mask[i] = 1
                if idx == 2:
                    number_mask[i] = 1
        except:
            word_mask = torch.zeros(512)
            number_mask = torch.zeros(512)
        
        word_word_mask = []
        number_word_mask = []
        for i, idx in enumerate(input_word_types):
            if idx == 3:
                word_word_mask += [1]
                number_word_mask += [0]
            if idx == 2:
                word_word_mask += [0]
                number_word_mask += [1]
                
        graphFactory = GraphFactory(question_tokens, table_tokens, paragraph_tokens, row_dict, column_dict, table_cell_types_flatten, table_pd, table_cell_type_ids, self.tokenizer)
        graph = graphFactory.generate_graph()
        
        return self._make_instance(input_ids, attention_mask,token_type_ids,
                question_tokens,
                table_pd, table_cell_type_ids,table_is_scale_type, table_number, table_number_index, table_cell_index, table_tokens,
                paragraph_tokens,paragraph_index,paragraph_number,paragraph_number_index, question_id,
                word_mask, number_mask, subword_lens, item_length, word_word_mask, number_word_mask,table_word_tokens,
                table_word_tokens_number, table_number_scale, paragraph_number_scale, graph, question,question_number, question_number_index)
        
    def _read(self, file_path: str):
        print("Reading file at %s", file_path)
        with open(file_path) as dataset_file:
            dataset = json.load(dataset_file)
        instances = []
        count = 0
        total = 0
        total_count = defaultdict(int)
        for one in tqdm(dataset):
            table = one['table']['table']
            if one['table']['uid'] == '097f22c33fd3ff811a21c799dd76e595':
                table = table[:-2]
            paragraphs = one['paragraphs']
            questions = one['questions']
            
            for question_answer in questions:
                question = question_answer["question"].strip()
                question_id = question_answer["uid"]
                instance = self._to_instance(question, table, paragraphs,question_id)
                if instance is not None:
                    instances.append(instance)
        print(len(instances))

        return instances
    



import io, requests, zipfile
import os
import json
import argparse
from datetime import datetime
from reg_hnt import options
import torch.nn as nn
from pprint import pprint
from reg_hnt.data.data_util import OPERATOR_CLASSES_
from reg_hnt.reghnt.util import create_logger, set_environment
from reg_hnt.reghnt.Vocabulary import *
from reg_hnt.data.tatqa_batch_gen import TaTQABatchGen
from transformers import RobertaModel, BertModel, ElectraModel, AutoModel
from reg_hnt.reghnt.modeling_reghnt import RegHNTModel
from reg_hnt.reghnt.model import RegHNTFineTuningModel
from pathlib import Path
import torch
parser = argparse.ArgumentParser("RegHNT training task.")
options.add_data_args(parser)
options.add_train_args(parser)
options.add_bert_args(parser)
parser.add_argument("--encoder", type=str, default='roberta_large')
parser.add_argument("--op_mode", type=int, default=0)
parser.add_argument("--ablation_mode", type=int, default=0)
parser.add_argument("--test_data_dir", type=str, default="./reg_hnt/cache")

args = parser.parse_args()
if args.ablation_mode != 0:
    args.save_dir = args.save_dir + "_{}_{}".format(args.op_mode, args.ablation_mode)
    args.data_dir = args.data_dir + "_{}_{}".format(args.op_mode, args.ablation_mode)

Path(args.save_dir).mkdir(parents=True, exist_ok=True)

args.cuda = args.gpu_num > 0
args_path = os.path.join(args.save_dir, "args.json")
with open(args_path, "w") as f:
    json.dump((vars(args)), f)

args.batch_size = args.batch_size // args.gradient_accumulation_steps

logger = create_logger("Roberta Training", log_file=os.path.join(args.save_dir, args.log_file))

pprint(args)
set_environment(args.seed, args.cuda)

def main():
    best_result = float("-inf")
    logger.info("Loading data...")

    train_itr = TaTQABatchGen(args, data_mode = "train", encoder=args.encoder)
    if args.ablation_mode != 3:
        dev_itr = TaTQABatchGen(args, data_mode="dev", encoder=args.encoder)
    else:
        dev_itr = TaTQABatchGen(args, data_mode="dev", encoder=args.encoder)

    num_train_steps = int(args.max_epoch * len(train_itr) / args.gradient_accumulation_steps)
    logger.info("Num update steps {}!".format(num_train_steps))

    logger.info(f"Build {args.encoder} model.")
    if args.encoder == 'bert':
        bert_model = BertModel.from_pretrained(args.bert_model)
    elif args.encoder == 'roberta_base':
        bert_model = RobertaModel.from_pretrained(args.roberta_model)
    elif args.encoder == 'roberta_large':
        bert_model = RobertaModel.from_pretrained(args.roberta_model)

    if args.ablation_mode == 0:
        operators = OPERATOR_CLASSES_

    if args.ablation_mode == 0:
        arithmetic_op_index = [3, 4, 6, 7, 8, 9]


    network = RegHNTModel(
        encoder = bert_model,
        config = bert_model.config,
        bsz = args.batch_size,
        operator_classes = len(OPERATOR_CLASSES),
        scale_classes = len(SCALE),
        operator_criterion = nn.CrossEntropyLoss(),
        scale_criterion = nn.CrossEntropyLoss(),
        arithmetic_op_index = arithmetic_op_index,
        op_mode = args.op_mode,
        ablation_mode = args.ablation_mode,
    )
    logger.info("Build optimizer etc...")

    model = RegHNTFineTuningModel(args, network, num_train_steps=num_train_steps)

    train_start = datetime.now()
    first = True
    for epoch in range(1, args.max_epoch + 1):
        model.reset()
        if not first:
            train_itr.reset()
        first = False
        logger.info('At epoch {}'.format(epoch))
        for step, batch in enumerate(train_itr):
            model.update(batch)
            if model.step % (args.log_per_updates * args.gradient_accumulation_steps) == 0 or model.step == 1:
                logger.info("Updates[{0:6}] train loss[{1:.5f}] remaining[{2}].\r\n".format(
                    model.updates, model.train_loss.avg,
                    str((datetime.now() - train_start) / (step + 1) * (num_train_steps - step - 1)).split('.')[0]))
                model.avg_reset()
        model.get_metrics(logger)

        model.reset()
        model.avg_reset()
        model.predict(dev_itr)
        metrics = model.get_metrics(logger)
        model.avg_reset()
        if metrics["f1"] > best_result:
            save_prefix = os.path.join(args.save_dir, "checkpoint_best")
            model.save(save_prefix, epoch)
            best_result = metrics["f1"]
            logger.info("Best eval F1 {} at epoch {}.\r\n".format(best_result, epoch))

if __name__ == "__main__":
    main()


import torch
import torch.nn as nn
import math

def gelu(x):
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))

class FFNLayer(nn.Module):
    def __init__(self, input_dim, intermediate_dim, output_dim, dropout, layer_norm=True):
        super(FFNLayer, self).__init__()
        self.fc1 = nn.Linear(input_dim, intermediate_dim)
        if layer_norm:
            self.ln = nn.LayerNorm(intermediate_dim)
        else:
            self.ln = None
        self.dropout_func = nn.Dropout(dropout)
        self.fc2 = nn.Linear(intermediate_dim, output_dim)

    def forward(self, input):

        inter = self.fc1(self.dropout_func(input))
        inter_act = gelu(inter)
        if self.ln:
            inter_act = self.ln(inter_act)
        return self.fc2(inter_act)    

class Retriever(nn.Module):
    def __init__(self, bert_model):
        super(Retriever, self).__init__()
        self.bert_model = bert_model
        self.hidden_size = self.bert_model.embeddings.word_embeddings.embedding_dim
        self.projection = FFNLayer(self.hidden_size, self.hidden_size, 1, 0.2)

    def forward(self, data):
        inputs = {"input_ids": data['input_ids'], "attention_mask": data['attention_mask']}
        cls_output = self.bert_model(**inputs)[0][:,0,:]
        logits = self.projection(cls_output)
        probs = torch.softmax(logits, 0)
        return probs
        


import torch
from classifier_module import ClassifierTrainer
MODEL_PATH= '/home/lfy/PTM/deberta-v3-large'

"""
CUDA_VISIBLE_DEVICES=0 python test_classifier.py \
--output_dir ./output \
--bert_model /home/lfy/PTM/deberta-v3-large \
--per_device_train_batch_size 4 --per_device_eval_batch_size 8 \
--data_path /home/lfy/lwh/data/mmqa
"""


trainer = ClassifierTrainer()
trainer.train()


import sys
import random
import torch
import numpy
import logging
from time import gmtime, strftime
from typing import Union, List, Tuple
from tatqa_eval import get_metrics, extract_gold_answers

def set_environment(seed, set_cuda=False):
    random.seed(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available() and set_cuda:
        torch.cuda.manual_seed_all(seed)


def create_logger(name, silent=False, to_disk=True, log_file=None):
    """Logger wrapper
    """
    # setup logger
    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG)
    log.propagate = False
    formatter = logging.Formatter(fmt='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S')
    if not silent:
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        log.addHandler(ch)
    if to_disk:
        log_file = log_file if log_file is not None else strftime("%Y-%m-%d-%H-%M-%S.log", gmtime())
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        log.addHandler(fh)
    return log


def metric_max_over_ground_truths(metric_fn, prediction, ground_truths):
    scores_for_ground_truths = []
    for ground_truth in ground_truths:
        score = metric_fn(prediction, ground_truth)
        scores_for_ground_truths.append(score)
    return max(scores_for_ground_truths)


class AverageMeter(object):
    """Computes and stores the average and current value."""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


class DropEmAndF1(object):
    """
    This :class:`Metric` takes the best span string computed by a model, along with the answer
    strings labeled in the data, and computes exact match and F1 score using the official DROP
    evaluator (which has special handling for numbers and for questions with multiple answer spans,
    among other things).
    """
    def __init__(self) -> None:
        self._total_em = 0.0
        self._total_f1 = 0.0
        self._count = 0

    def __call__(self, prediction: Union[str, List], ground_truth: List[str]):  # type: ignore
        """
        Parameters
        ----------
        prediction: ``Union[str, List]``
            The predicted answer from the model evaluated. This could be a string, or a list of string
            when multiple spans are predicted as answer.
        ground_truths: ``dict``
            All the ground truth answer annotations.
        """
        # If you wanted to split this out by answer type, you could look at [1] here and group by
        # that, instead of only keeping [0].
        # ground_truth_answer_strings = [answer_json_to_strings(annotation)[0] for annotation in ground_truths]
        #  gold_answers is a string or list of string
        # gold_type, gold_answers, gold_unit = extract_gold_answers(ground_truth)

        if not prediction:
            self._count += 1
        else:
            ground_truth_answer_strings = ground_truth # a list of list
            # print(ground_truth_answer_strings)
            # print(prediction)
            exact_match, f1_score = metric_max_over_ground_truths(
                    get_metrics,
                    [prediction],
                    [ground_truth_answer_strings]
            )
            self._total_em += exact_match
            self._total_f1 += f1_score
            self._count += 1

    def get_metric(self, reset: bool = False) -> Tuple[float, float]:
        """
        Returns
        -------
        Average exact match and F1 score (in that order) as computed by the official DROP script
        over all inputs.
        """
        exact_match = self._total_em / self._count if self._count > 0 else 0
        f1_score = self._total_f1 / self._count if self._count > 0 else 0
        if reset:
            self.reset()
        return exact_match, f1_score

    def reset(self):
        self._total_em = 0.0
        self._total_f1 = 0.0
        self._count = 0

    def __str__(self):
        return f"DropEmAndF1(em={self._total_em}, f1={self._total_f1})"


import sys
sys.path.append('./')
import json
import os
import math
from tqdm import tqdm
import torch.nn.functional as F
import torch.nn as nn
from typing import List, Dict
from .utils import *
from .dataset import ReadDataset, BartReadDataset
from transformers import AutoTokenizer, AutoModel, BartForConditionalGeneration
from torch.utils.data import Dataset, DataLoader
from transformers import get_linear_schedule_with_warmup, AdamW
import numpy as np
from pathlib import Path
from .evaluate_script1 import get_raw_scores
from .model import ReadModel


class HQATrainer:
    def __init__(self, args):
        self.args = args
    
        Path(args.output_path).mkdir(parents=True, exist_ok=True)
        self.logger = create_logger("ReadModel", log_file=os.path.join(args.output_path, 'log.txt'))
        self.tokenizer = AutoTokenizer.from_pretrained(args.plm)
        with open(args.dev_reference, 'r') as f:
            self.reference = json.load(f)

        if args.is_train:
            self.train_data = load_json(args.train_data_path)
            self.logger.info(f"train_data: {len(self.train_data)}")
        self.dev_data = load_json(args.dev_data_path)
        self.logger.info(f"dev_data: {len(self.dev_data)}")
    @staticmethod
    def collate(self, data, **training_config):
        pass
    @staticmethod
    def train(self, data, **training_config):
        pass
    @staticmethod
    def eval(self, model, loader):
        pass
    @staticmethod
    def train_epoch(self, loader, model):
        pass
    
class GenerateTrainer(HQATrainer):
    def __init__(self, args):
        super(GenerateTrainer, self).__init__(args)
        if args.is_train:
            self.train_dataset = BartReadDataset(args, self.train_data, 'train')
        self.dev_dataset = BartReadDataset(args, self.dev_data, 'dev')
        
    def collate(self, data, **training_config):
        bs = len(data)
        max_input_length = 0
        input_ids = []
        answer_ids = []
        metadata = []
        max_input_length = max([len(item['input_ids']) for item in data])
        max_answer_length = max([len(item['answer_ids']) for item in data])
        if max_input_length > training_config['max_len']:
            max_input_length = training_config['max_len']

        for i in range(bs):
            if len(data[i]['input_ids']) > max_input_length:
                input_id = data[i]['input_ids'][:max_input_length]
            else:
                input_id = data[i]['input_ids'] + (max_input_length - len(data[i]['input_ids'])) * [self.tokenizer.pad_token_id]
            input_ids.append(input_id)


            if len(data[i]['answer_ids']) > max_answer_length:
                answer_id = data[i]['answer_ids'][:max_answer_length]
            else:
                answer_id = data[i]['answer_ids'] + (max_answer_length - len(data[i]['answer_ids'])) * [self.tokenizer.pad_token_id]
            answer_ids.append(answer_id)

            metadata.append(data[i])
        input_ids = torch.tensor(input_ids)
        input_mask = torch.where(input_ids==self.tokenizer.pad_token_id, 0, 1)

        answer_ids = torch.tensor(answer_ids)
        answer_mask = torch.where(answer_ids==self.tokenizer.pad_token_id, 0, 1)

        return {"input_ids": input_ids.cuda(), "input_mask":input_mask.cuda(), "answer_ids":answer_ids.cuda(), "answer_mask":answer_mask.cuda(), "metadata":metadata}

    def train(self, **training_config):
        self.training_config = training_config
        
        train_loader = DataLoader(self.train_dataset, batch_size=training_config['train_bs'], collate_fn=lambda x: self.collate(x, **training_config))
        dev_loader = DataLoader(self.dev_dataset, batch_size=training_config['dev_bs'], collate_fn=lambda x: self.collate(x, **training_config))

        model = BartForConditionalGeneration.from_pretrained(self.args.plm)
        device = torch.device("cuda")
        model.to(device)
        
        best_acc = -1
        self.optimizer = AdamW(model.parameters(), lr=training_config['lr'], eps=training_config['eps'])
        t_total = len(self.train_dataset) * training_config['epoch_num']
        self.scheduler = get_linear_schedule_with_warmup(self.optimizer, num_warmup_steps=0 * t_total, num_training_steps=t_total)

        for epoch in range(training_config['epoch_num']):
            self.logger.info(f"Training epoch: {epoch}")
            self.train_epoch(train_loader, model)
            self.logger.info(f"start eval....")
            total_exact, outputs = self.eval(model, dev_loader)
            scores = get_raw_scores(outputs, self.reference)
            total_exact = scores['total exact']
            self.logger.info(f"{scores}")
            self.logger.info(f"Total Exact: {total_exact}")
            if total_exact > best_acc:
                best_acc = total_exact
                model_to_save = model.module if hasattr(model, "module") else model
                model_save_path = os.path.join(self.args.output_path, 'ckpt.pt')
                self.logger.info(f"saving model...to {model_save_path}")
                torch.save(model_to_save.state_dict(), model_save_path)
            
    
    def eval(self, model, loader):
        model.eval()
        total, acc = 0, 0
        outputs = []
        with torch.no_grad():
            for i, data in enumerate(tqdm(loader)):
                generated_ids = model.generate(
                    input_ids = data['input_ids'],
                    attention_mask=data['input_mask'],
                    max_length = 20,
                    num_beams = 3,
                    early_stopping=True
                )
                metadata = data['metadata']
                for i, item in enumerate(generated_ids):
                    total += 1
                    output = torch.masked_select(item, item.ge(3))
                    pred_answer = self.tokenizer.decode(output)
                    answer = metadata[i]['answer-text']
                    question_id = metadata[i]['question_id']
                    if pred_answer == answer:
                        acc += 1
                    outputs.append({"question_id":question_id, "pred":pred_answer})
        self.logger.info(f"outputs num: {len(outputs)}")
        return acc / total, outputs
    
            
            
    def train_epoch(self, loader, model):
        model.train()
        averge_step = len(loader) // 12
        loss_sum, step = 0, 0
        for i, data in enumerate(tqdm(loader)):
            labels = data['answer_ids'][:, 1:].clone()
            labels[labels==self.tokenizer.pad_token_id] = -100
            outputs = model(input_ids=data['input_ids'], 
                            attention_mask=data['input_mask'], 
                            decoder_input_ids=data['answer_ids'][:, :-1],
                            labels=labels)
            loss = outputs[0]
            self.optimizer.zero_grad()
            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), 0.1)
            self.optimizer.step()
            self.scheduler.step()
            loss_sum += loss
            step += 1
            if i % averge_step == 0:
                self.logger.info("Training Loss [{0:.5f}]".format(loss_sum/step))
                loss_sum, step = 0, 0
    

class MRCTrainer(HQATrainer):
    def __init__(self, args):
        super(MRCTrainer, self).__init__(args)
        if args.is_train:
            self.train_dataset = ReadDataset(args, self.train_data[:100], 'train')
        self.dev_dataset = ReadDataset(args, self.dev_data[:100], 'dev')
    def collate(self, data, **training_config):
        bs = len(data)
        max_input_length = 0
        input_ids = []
        metadata = []
        start_labels, end_labels = [], []
        max_input_length = max([len(item['input_ids']) for item in data])
        if max_input_length > training_config['max_len']:
            max_input_length = training_config['max_len']
        for i in range(bs):
            if len(data[i]['input_ids']) > max_input_length:
                input_id = data[i]['input_ids'][:max_input_length]
            else:
                input_id = data[i]['input_ids'] + (max_input_length - len(data[i]['input_ids'])) * [self.tokenizer.pad_token_id]
            input_ids.append(input_id)
            start_labels.append(data[i]['start'])
            end_labels.append(data[i]['end'])
            metadata.append(data[i]['metadata'])
        input_ids = torch.tensor(input_ids)
        input_mask = torch.where(input_ids==self.tokenizer.pad_token_id, 0, 1)
        start_labels = torch.tensor(start_labels)
        end_labels = torch.tensor(end_labels)
        return {"input_ids": input_ids.cuda(), "input_mask":input_mask.cuda(), "start_labels":start_labels.cuda(), "end_labels":end_labels.cuda(), "metadata":metadata}

    def train(self, **training_config):
        self.training_config = training_config
        
        train_loader = DataLoader(self.train_dataset, batch_size=training_config['train_bs'], collate_fn=lambda x: self.collate(x, **training_config))
        dev_loader = DataLoader(self.dev_dataset, batch_size=training_config['dev_bs'], collate_fn=lambda x: self.collate(x, **training_config))

        bert_model = AutoModel.from_pretrained(self.args.plm)
        model = ReadModel(bert_model)
        device = torch.device("cuda")
        model.to(device)
        
        best_acc = -1
        self.optimizer = AdamW(model.parameters(), lr=training_config['lr'], eps=training_config['eps'])
        t_total = len(self.train_dataset) * training_config['epoch_num']
        self.scheduler = get_linear_schedule_with_warmup(self.optimizer, num_warmup_steps=0 * t_total, num_training_steps=t_total)

        for epoch in range(training_config['epoch_num']):
            self.logger.info(f"Training epoch: {epoch}")
            self.train_epoch(train_loader, model)
            self.logger.info(f"start eval....")
            _, outputs = self.eval(model, dev_loader)
            scores = get_raw_scores(outputs, self.reference)
            total_exact = scores['total exact']
            self.logger.info(f"{scores}")
            self.logger.info(f"Total Exact: {total_exact}")
            if total_exact > best_acc:
                best_acc = total_exact
                model_to_save = model.module if hasattr(model, "module") else model
                model_save_path = os.path.join(self.args.output_path, 'ckpt.pt')
                self.logger.info(f"saving model...to {model_save_path}")
                torch.save(model_to_save.state_dict(), model_save_path)


    def train_epoch(self, loader, model):
        model.train()
        averge_step = len(loader) // 8
        loss_sum, step = 0, 0
        for it, data in enumerate(tqdm(loader)):
            start_logits, end_logits = model(data)
            loss_func = nn.CrossEntropyLoss(reduction='sum')
            start_loss = loss_func(start_logits, data['start_labels'])
            end_loss = loss_func(end_logits, data['end_labels'])
            loss = (start_loss + end_loss) / 2

            self.optimizer.zero_grad()
            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), 0.1)
            self.optimizer.step()
            self.scheduler.step()
            loss_sum += loss
            step += 1
            if it % averge_step == 0:
                self.logger.info("Training Loss [{0:.5f}]".format(loss_sum/step))
                loss_sum, step = 0, 0

    def eval(self, model, loader):
        model.eval()
        total, acc = 0, 0
        outputs = []
        with torch.no_grad():
            for i, data in enumerate(tqdm(loader)):
                start_logits, end_logits = model(data)
                start_pred = torch.argmax(F.softmax(start_logits,dim=1), dim=1).cpu().tolist()
                end_pred = torch.argmax(F.softmax(end_logits,dim=1), dim=1).cpu().tolist()
                metadata = data['metadata']
                for i in range(len(metadata)):
                    total += 1
                    question_id = metadata[i]['question_id']
                    if start_pred[i] > end_pred[i]:
                        end_pred[i] = start_pred[i]

                    answer = metadata[i]['answer-text']

                    ids_token_index = metadata[i]['ids_token_index']
                    tokens = metadata[i]['tokens']

                    try:
                        start_token_index = ids_token_index[start_pred[i]]
                        end_token_index = ids_token_index[end_pred[i]]
                    except:
                        continue
                    pred_answer = ' '.join(tokens[start_token_index:end_token_index+1])
                    if pred_answer == answer:
                        acc += 1
                    outputs.append({"question_id":question_id, "pred":pred_answer})
        print(f"outputs num: {len(outputs)}")
        return acc / total, outputs
        
    


import os
import json
import numpy as np
from dataclasses import dataclass
from typing import Dict, List, Sequence, Tuple, Union

from transformers.trainer import PredictionOutput
from transformers.tokenization_utils import PreTrainedTokenizer
import jieba
from rouge_chinese import Rouge
from nltk.translate.bleu_score import sentence_bleu, SmoothingFunction
from .peft_trainer import PeftTrainer
from .utils import get_logger, IGNORE_INDEX


logger = get_logger(__name__)


@dataclass
class ComputeMetrics:
    r"""
    Wraps the tokenizer into metric functions, used in Seq2SeqPeftTrainer.
    """

    tokenizer: PreTrainedTokenizer

    def __call__(self, eval_preds: Sequence[Union[np.ndarray, Tuple[np.ndarray]]]) -> Dict[str, float]:
        r"""
        Uses the model predictions to compute metrics.
        """
        preds, labels = eval_preds
        score_dict = {"rouge-1": [], "rouge-2": [], "rouge-l": [], "bleu-4": []}

        for pred, label in zip(preds, labels):
            pred_pad_len, label_pad_len = np.sum(pred == IGNORE_INDEX), np.sum(label == IGNORE_INDEX)
            pred = pred[len(label) - label_pad_len : len(pred) - pred_pad_len] # remove prompts
            label = label[:len(label) - label_pad_len]

            hypothesis = list(jieba.cut(self.tokenizer.decode(pred, skip_special_tokens=True)))
            reference = list(jieba.cut(self.tokenizer.decode(label, skip_special_tokens=True)))

            if len(" ".join(hypothesis).split()) == 0:
                result = {"rouge-1": {"f": 0.0}, "rouge-2": {"f": 0.0}, "rouge-l": {"f": 0.0}}
            else:
                rouge = Rouge()
                scores = rouge.get_scores(" ".join(hypothesis), " ".join(reference))
                result = scores[0]

            for k, v in result.items():
                score_dict[k].append(round(v["f"] * 100, 4))

            bleu_score = sentence_bleu([list(label)], list(pred), smoothing_function=SmoothingFunction().method3)
            score_dict["bleu-4"].append(round(bleu_score * 100, 4))

        return {k: float(np.mean(v)) for k, v in score_dict.items()}


class Seq2SeqPeftTrainer(PeftTrainer):
    r"""
    Inherits PeftTrainer to compute generative metrics such as BLEU and ROUGE.
    """

    def save_predictions(
            self,
            predict_results: PredictionOutput
    ) -> None:
        r"""
        Saves model predictions to `output_dir`.

        A custom behavior that not contained in Seq2SeqTrainer.
        """
        if not self.is_world_process_zero():
            return

        output_prediction_file = os.path.join(self.args.output_dir, "generated_predictions.jsonl")
        logger.info(f"Saving prediction results to {output_prediction_file}")
        with open(output_prediction_file, "w", encoding="utf-8") as writer:
            res: List[str] = []
            for pred, label in zip(predict_results.predictions, predict_results.label_ids):
                pred_pad_len, label_pad_len = np.sum(pred == IGNORE_INDEX), np.sum(label == IGNORE_INDEX)
                pred = pred[len(label) - label_pad_len : len(pred) - pred_pad_len] # remove prompts
                label = label[:len(label) - label_pad_len]

                pred = self.tokenizer.decode(pred, skip_special_tokens=True)
                label = self.tokenizer.decode(label, skip_special_tokens=True)

                res.append(json.dumps({"label": label, "predict": pred}, ensure_ascii=False))

            writer.write("\n".join(res))



import os
import platform

import re
import logging
import json
from structs.data import Cell, Table, HFTabularDataset


logger = logging.getLogger(__name__)

class SpreadSheetQA(HFTabularDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _load_split(self, split):
        question_file_map = {
            "train": "train.json",
            "dev": "dev.json",
            "test": "test.json"
        }

        file_base_path = 'datasets/SpreadSheetQA'
        logger.info(f"Loading SpreadSheetQA - {split}")
        question_file_path = os.path.join(file_base_path, question_file_map[split])

        with open(question_file_path, 'r', encoding='utf-8') as f:
            question_data_list = json.load(f)

        data = []

        for question_data in question_data_list:

            properties_info = {}

            if 'explanation' in question_data['qa']:
                properties_info['question--explanation'] = str(question_data['qa']['explanation'])
            if 'program' in question_data['qa']:
                properties_info['question--program'] = str(question_data['qa']['program'])
            if 'program_re' in question_data['qa']:
                properties_info['question--program_re'] = str(question_data['qa']['program_re'])
            if 'source' in question_data:
                properties_info['source'] = str(question_data['source'])

            question = question_data['qa']['question']
            if type(question_data['table_ori']) == list:
                table_data = question_data['table_ori']
            else:
                table_data = question_data['table_ori']['table']
            table_data_headers = table_data[0]
            table_data_contents = table_data[1:]
            txt_list = question_data['pre_text'] + question_data['post_text']
            txt_list = list(filter(lambda x: x not in ['.', '*'], txt_list))


            data.append({
                "question": question,
                "table": {
                    "header": table_data_headers,
                    "content": table_data_contents
                },
                "txt": txt_list,
                "properties": properties_info
            })
        self.data[split] = data
        self.dataset_info = {
            "citation": "111",
            "description": "111",
            "version": "111",
            "license": "111"
        }

    def load_split_test(self, split):
        self._load_split(split)

    def prepare_table(self, entry):
        t = Table()
        t.type = "default"
        t.default_question = entry["question"]

        t.props = entry['properties']
        t.txt_info = entry['txt']

        for header_cell in entry["table"]["header"]:
            c = Cell()
            c.value = header_cell
            c.is_col_header = True
            t.add_cell(c)
        t.save_row()

        for row in entry["table"]["content"]:
            for (i, cell) in enumerate(row):
                c = Cell()
                if i == 0:
                    c.is_row_header = True
                c.value = cell
                t.add_cell(c)
            t.save_row()
        return t
if __name__ == '__main__':
    SpreadSheetQA().load_split_test("train")


from typing import Set, Tuple, Union
from tatqa_utils import *
import pandas as pd
import numpy as np
from scipy.optimize import linear_sum_assignment


def _answer_to_bags(answer: Union[str, List[str], Tuple[str, ...]]) -> Tuple[List[str], List[Set[str]]]:
    if isinstance(answer, (list, tuple)):
        raw_spans = answer
    else:
        raw_spans = [answer]
    normalized_spans: List[str] = []
    token_bags = []
    for raw_span in raw_spans:
        normalized_span = normalize_answer(raw_span)
        normalized_spans.append(normalized_span)
        token_bags.append(set(normalized_span.split()))
    return normalized_spans, token_bags


def _align_bags(predicted: List[Set[str]], gold: List[Set[str]]) -> List[float]:
    """
    Takes gold and predicted answer sets and first finds the optimal 1-1 alignment
    between them and gets maximum metric values over all the answers.
    """
    scores = np.zeros([len(gold), len(predicted)])
    for gold_index, gold_item in enumerate(gold):
        for pred_index, pred_item in enumerate(predicted):
            # if _match_numbers_if_present(gold_item, pred_item): no need to match number in tatqa
            scores[gold_index, pred_index] = _compute_f1(pred_item, gold_item)
    row_ind, col_ind = linear_sum_assignment(-scores)

    max_scores = np.zeros([max(len(gold), len(predicted))])
    for row, column in zip(row_ind, col_ind):
        max_scores[row] = max(max_scores[row], scores[row, column])
    return max_scores


def _compute_f1(predicted_bag: Set[str], gold_bag: Set[str]) -> float:
    intersection = len(gold_bag.intersection(predicted_bag))
    if not predicted_bag:
        precision = 1.0
    else:
        precision = intersection / float(len(predicted_bag))
    if not gold_bag:
        recall = 1.0
    else:
        recall = intersection / float(len(gold_bag))
    f1 = (2 * precision * recall) / (precision + recall) if not (precision == 0.0 and recall == 0.0) else 0.0
    return f1


def _match_numbers_if_present(gold_bag: Set[str], predicted_bag: Set[str]) -> bool:
    gold_numbers = set()
    predicted_numbers = set()
    for word in gold_bag:
        if is_number(word):
            gold_numbers.add(word)
    for word in predicted_bag:
        if is_number(word):
            predicted_numbers.add(word)
    if (not gold_numbers) or gold_numbers.intersection(predicted_numbers):
        return True
    return False


def get_metrics(predicted: Union[str, List[str], Tuple[str, ...]],
                gold: Union[str, List[str], Tuple[str, ...]]) -> Tuple[float, float]:
    """
    Takes a predicted answer and a gold answer (that are both either a string or a list of
    strings), and returns exact match and the DROP F1 metric for the prediction.  If you are
    writing a script for evaluating objects in memory (say, the output of predictions during
    validation, or while training), this is the function you want to call, after using
    :func:`answer_json_to_strings` when reading the gold answer from the released data file.
    """
    predicted_bags = _answer_to_bags(predicted)
    gold_bags = _answer_to_bags(gold)

    if set(predicted_bags[0]) == set(gold_bags[0]) and len(predicted_bags[0]) == len(gold_bags[0]):
        exact_match = 1.0
    else:
        exact_match = 0.0

    f1_per_bag = _align_bags(predicted_bags[1], gold_bags[1])
    f1 = np.mean(f1_per_bag)
    f1 = round(f1, 2)
    return exact_match, f1


def extract_gold_answers(qa_annotation):
    '''
    span
    multi-span
    arithmetic (+ - * /)
    count
    date
    other
    gold answers is a list of list, each item in gold answers is a valid answer
    '''
    answer_type, scale = qa_annotation["answer_type"], qa_annotation['scale']
    answer_content = qa_annotation['answer']
    gold_answers = []
    if answer_type in ['multi-span', 'span']: # list
        assert isinstance(answer_content, list), answer_content
        gold_answers = answer_content # multi-span
    elif answer_type in ["arithmetic"]:
        gold_answers.append(str(answer_content))
    elif answer_type in ['count']:
        gold_answers.append(str(int(answer_content)))
    else:
        gold_answers.append(str(answer_content))
    return answer_type, gold_answers, scale


def metric_max_over_ground_truths(metric_fn, predictions, ground_truths):
    scores_for_ground_truths = []
    for pred in predictions:
        for ground_truth in ground_truths:
            score = metric_fn(pred, ground_truth)
            scores_for_ground_truths.append(score)
    if len(scores_for_ground_truths) == 0:
        return 0, 0
    return max(scores_for_ground_truths)


class AverageMeter(object):
    """Computes and stores the average and current value."""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def get_answer_str(answers: list, scale: str):
    """
    :param ans_type:  span, multi-span, arithmetic, count
    :param ans_list:
    :param scale: "", thousand, million, billion, percent
    :param mode:
    :return:

    """
    sorted_ans = sorted(answers)
    ans_temp = []
    for ans in sorted_ans:
        ans_str = str(ans)
        if is_number(ans_str):
            ans_num = to_number(ans_str)
            if ans_num is None:
                if scale:
                    ans_str = ans_str + " " + str(scale)
            else:
                if '%' in ans_str: #  has been handled the answer itself is a percentage
                    ans_str = '%.4f' % ans_num
                else:
                    ans_str = '%.4f' % (round(ans_num, 2) * scale_to_num(scale))
        else:
            if scale:
                ans_str = ans_str + " " + str(scale)
        ans_temp.append(ans_str)
    return [" ".join(ans_temp)]


# handle percentage
def add_percent_pred(prediction_strings, pred_scale, pred):
    """
    to solve [pred = 0.2342] <>   [ans = 23.42 and scale == 'percent']

    :param prediction_strings:
    :param gold_ans_type:
    :param gold_scale:
    :param pred:
    :return:
    """
    if len(pred) > 1:
        return prediction_strings
    pred_str = str(pred[0])
    if pred_str is None:
        return prediction_strings
    if not pred_scale and '%' not in pred_str and is_number(pred_str): # mode only or no pred_scale num only
        pred_str = to_number(pred_str)
        if pred_str is None:
            return prediction_strings
        prediction_strings.append('%.4f' % pred_str)
    return prediction_strings


class TaTQAEmAndF1(object):
    """
    This :class:`Metric` takes the best span string computed by a model, along with the answer
    strings labeled in the data, and computes exact match and F1 score using the official DROP
    evaluator (which has special handling for numbers and for questions with multiple answer spans,
    among other things).
    """
    def __init__(self) -> None:
        self._total_em = 0.0
        self._total_f1 = 0.0
        self._scale_em = 0.0
        self._op_em = 0.0
        self.op_correct_count = {"SPAN": 0, "MULTI_SPAN": 1, "COUNT": 2, "ARITHMETIC": 3}
        self.op_total_count = {"SPAN": 0, "MULTI_SPAN": 1, "COUNT": 2, "ARITHMETIC": 3}
        self.scale_correct_count = {"": 0, "thousand": 0, "million": 0, "billion": 0, "percent": 0}
        self.scale_total_count = {"": 0, "thousand": 0, "million": 0, "billion": 0, "percent": 0}
        self._count = 0
        self._details = []

    def __call__(self,
                 ground_truth: dict,
                 prediction: Union[str, List],
                 pred_scale="",
                 pred_span = None,
                 gold_span = None,
                 pred_op=None,
                 gold_op=None):  # type: ignore
        """
        :param ground_truth:
        :param prediction:
        :param pred_scale:
        :param pred_span:
        :param gold_span:
        :param pred_op:
        :param gold_op:
        :return:
        """
        if pred_op is not None:
            if pred_op == 'ARITHMETIC' and gold_op == 'ARITHMETIC':
                self.op_correct_count[pred_op] += 1
                self._op_em += 1
            elif pred_op in ['SPAN', 'COUNT', 'MULTI_SPAN'] and gold_op in ['SPAN', 'COUNT', 'MULTI_SPAN']:
                self.op_correct_count[pred_op] += 1
                self._op_em += 1
            self.op_total_count[gold_op] += 1
        if pred_scale == ground_truth["scale"]:
            self.scale_correct_count[pred_scale] += 1
        self.scale_total_count[ground_truth["scale"]] += 1
        if prediction is None or prediction == []:
            exact_match = 0
            f1_score = 0
            span_exact_match = 0
            span_f1_score = 0
        else:
            gold_type, gold_answer, gold_scale = extract_gold_answers(ground_truth)
            if gold_answer is None:
                exact_match = 0
                f1_score = 0
                span_exact_match = 0
                span_f1_score = 0
            else:
                ground_truth_answer_strings = get_answer_str(gold_answer, gold_scale)

                if gold_scale == pred_scale:
                    self._scale_em += 1
                prediction = prediction if isinstance(prediction, list) else [prediction]
                prediction_strings = get_answer_str(prediction, pred_scale)
                prediction_strings = add_percent_pred(prediction_strings, pred_scale, prediction)
                exact_match, f1_score = metric_max_over_ground_truths(
                        get_metrics,
                        prediction_strings,
                        ground_truth_answer_strings
                )
                
                # if ground_truth['uid'] == 'adfc652018f8560af945f804f8557ab8':
                #     import pdb; pdb.set_trace()
                # if ground_truth['answer_from'] == 'text' and exact_match != 1:
                #     import pdb; pdb.set_trace()
                # gold_answer, prediction, exact_match, ground_truth['uid']
                if gold_type in ['arithmetic', 'count']:
                    """if gold type equals with arithmetic and count, set the f1_score == exact_match"""
                    f1_score = exact_match
                if not pred_span:
                    span_exact_match = 0
                    span_f1_score = 0
                else:
                    pred_span_strings = get_answer_str(pred_span, "")
                    gold_span_strings = get_answer_str(gold_span, "")
                    span_exact_match, span_f1_score = metric_max_over_ground_truths(
                        get_metrics,
                        pred_span_strings,
                        gold_span_strings,
                    )

        self._total_em += exact_match
        self._total_f1 += f1_score
        self._count += 1
        it = {**ground_truth,
              **{"pred":prediction,
                 "pred_scale":pred_scale,
                 "em":exact_match,
                 "f1":f1_score,
                 "pred_span":pred_span,
                 "gold_span":gold_span,
                 "span_em":span_exact_match,
                 "span_f1":span_f1_score}}
        self._details.append(it)
        return exact_match

    def get_overall_metric(self, reset: bool = False) -> Tuple[float, float, float, float]:
        """
        Returns
        -------
        Average exact match and F1 score (in that order) as computed by the official DROP script
        over all inputs.
        """
        exact_match = self._total_em / self._count if self._count > 0 else 0
        f1_score = self._total_f1 / self._count if self._count > 0 else 0
        scale_score = self._scale_em / self._count if self._count > 0 else 0
        op_score = self._op_em / self._count if self._count > 0 else 0
        op_em_detail = {"SPAN": 0, "MULTI_SPAN": 1, "COUNT": 2, "ARITHMETIC": 3}
        scale_em_detail = {"": 0, "thousand": 0, "million": 0, "billion": 0, "percent": 0}
        for k in op_em_detail.keys():
            op_em_detail[k] = self.op_correct_count[k] / self.op_total_count[k] if self.op_total_count[k] > 0 else 0

        for k in scale_em_detail.keys():
            scale_em_detail[k] = self.scale_correct_count[k] / self.scale_total_count[k] if self.scale_total_count[k] > 0 else 0
        if reset:
            self.reset()
        return exact_match, f1_score, scale_score, op_score

    def get_detail_metric(self):
        df = pd.DataFrame(self._details)
        if len(self._details) == 0:
            return None, None
        em_pivot_tab = df.pivot_table(index='answer_type', values=['em'],
                                    columns=['answer_from'], aggfunc='mean').fillna(0)

        f1_pivot_tab = df.pivot_table(index='answer_type', values=['f1'],
                                    columns=['answer_from'], aggfunc='mean').fillna(0)
        return em_pivot_tab, f1_pivot_tab

    def get_raw_pivot_table(self):
        df = pd.DataFrame(self._details)
        pivot_tab = df.pivot_table(index='answer_type', values=['em'],
                                  columns=['answer_from'], aggfunc='count').fillna(0)
        return pivot_tab

    def get_raw(self):
        return self._details

    def reset(self):
        self._total_em = 0.0
        self._total_f1 = 0.0
        self._scale_em = 0.0
        self._op_em = 0.0
        self._count = 0
        self._details = []
        self.op_correct_count = {"SPAN": 0, "MULTI_SPAN": 1, "COUNT": 2, "ARITHMETIC": 3}
        self.op_total_count = {"SPAN": 0, "MULTI_SPAN": 1, "COUNT": 2, "ARITHMETIC": 3}
        self.scale_correct_count = {"": 0, "thousand": 0, "million": 0, "billion": 0, "percent": 0}
        self.scale_total_count = {"": 0, "thousand": 0, "million": 0, "billion": 0, "percent": 0}

    def __str__(self):
        return f"TaTQAEmAndF1(em={self._total_em}, f1={self._total_f1}, count={self._count})"


import torch
from retriever_module import RetrieverTrainer

"""

CUDA_VISIBLE_DEVICES=0,1,2 python test_retriever.py \
--output_dir ./output \
--bert_model /home/lfy/PTM/deberta-v3-large \
--data_path /home/lfy/lwh/data/mmqa \
--n_gpu 3 \
--image_or_text image \
--data_path /home/lfy/lwh/data/mmqa
"""

trainer = RetrieverTrainer()
trainer.train()


import json
from abc import ABC, abstractmethod
from typing import List, Optional, Dict


class GPTDataSet(ABC):
    def __init__(self, data_path: str, demo_path: Optional[str]):
        self.data_path = data_path
        self.demo_path = demo_path
        self.data = self.read_data(self.data_path)
        self.demos = self.read_demos(self.demo_path)

    @abstractmethod
    def read_data(self, data_path: str) -> List[Dict[str, any]]:
        """
        Read the data from the specified data path and preprocess it into the following format:
        :param data_path: The path of the data file.
        :return: A list of dictionaries, each containing the following keys:
            - "id": The unique identifier of the data.
            - "question": The question associated with the data.
            - "texts": A list of strings representing the texts associated with the data.
            - "rows": A list representing the tables' rows associated with the data.
        """
        pass

    @abstractmethod
    def read_demos(self, demo_path: str) -> List[Dict[str, any]]:
        """
        Read the demonstration data from the specified demo path and preprocess it into the following format:
        :param demo_path: The path of the demonstration file.
        :return: A list of dictionaries, each containing the following keys:
            - "question": The question of the demonstration.
            - "rationale": The CoT (Causes of Truth) rationale of the demonstration.
            - "answer": The golden answer of the demonstration.
        """
        pass


class MultiHiertt(GPTDataSet):
    def read_data(self, data_path: str) -> List[Dict]:
        data_list = []
        data = json.load(
            open(data_path, 'r', encoding='utf-8')
        )
        for one in data:
            texts = []
            rows = []
            for text_evidence in one["qa"]["text_evidence"]:
                texts.append(one["paragraphs"][text_evidence])

            for table_evidence in one["qa"]["table_evidence"]:
                rows.append(one["table_description"][table_evidence])
            data_list.append({
                "id": one["uid"],
                "question": one["qa"]["question"],
                "texts": texts,
                "rows": rows,
            })
        return data_list

    def read_demos(self, demo_path: str) -> List[Dict]:
        if demo_path is not None:
            return json.load(
                open(demo_path, 'r', encoding='utf-8')
            )
        else:
            return []


class UnifiedSKG(GPTDataSet):
    def read_data(self, data_path: str) -> List[Dict[str, any]]:
        data_list = []
        data = json.load(
            open(data_path, 'r', encoding='utf-8')
        )
        for one in data:
            data_list.append({
                "id": self.get_id(one),
                "question": one["question"],
                "texts": None,
                "rows": self.get_rows(one),
            })
        return data_list

    @abstractmethod
    def get_id(self, one) -> Optional[str]:
        pass

    @abstractmethod
    def get_rows(self, one) -> List[str]:
        pass

    def read_demos(self, demo_path: str) -> List[Dict[str, any]]:
        if demo_path is not None:
            return json.load(
                open(demo_path, 'r', encoding='utf-8')
            )
        else:
            return []


class Wikisql(UnifiedSKG):
    def get_id(self, one) -> Optional[str]:
        return None

    def get_rows(self, one) -> List[str]:
        return [one["table"]["header"]] + one["table"]["rows"]


class Wikitq(UnifiedSKG):
    def get_id(self, one) -> Optional[str]:
        return one["id"]

    def get_rows(self, one) -> List[str]:
        return [one["table"]["header"]] + one["table"]["rows"]


class FinQA(GPTDataSet):
    def read_data(self, data_path: str) -> List[Dict[str, any]]:
        data_list = []
        data = json.load(
            open(data_path, 'r', encoding='utf-8')
        )
        for one in data:
            texts = []
            for k, v in one["qa"]["gold_inds"].items():
                if "text" in k:
                    texts.append(v)
            data_list.append({
                "id": one["id"],
                "question": one["qa"]["question"],
                "texts": texts,
                "rows": one["table"],
            })
        return data_list

    def read_demos(self, demo_path: str) -> List[Dict[str, any]]:
        if demo_path is not None:
            return json.load(
                open(demo_path, 'r', encoding='utf-8')
            )
        else:
            return []


class TAT_QA(GPTDataSet):
    def read_data(self, data_path: str) -> List[Dict[str, any]]:
        data_list = []
        data = json.load(
            open(data_path, 'r', encoding='utf-8')
        )
        for one in data:
            rows = one["table"]["table"]
            texts = []
            for p in one["paragraphs"]:
                texts.append(p["text"])
            for question in one["questions"]:
                data_list.append({
                    "id": question["uid"],
                    "question": question["question"],
                    "texts": texts,
                    "rows": rows,
                })
        return data_list

    def read_demos(self, demo_path: str) -> List[Dict[str, any]]:
        if demo_path is not None:
            return json.load(
                open(demo_path, 'r', encoding='utf-8')
            )
        else:
            return []



from loaders.WikiSQL import WikiSQL
from loaders.WikiTableQuestions import WikiTableQuestion
from loaders.TATQA import TATQA
from loaders.FinQA import FinQA
from loaders.HiTab import HiTab
from loaders.HybridQA import HybridQA
from loaders.MultiHiertt import MultiHiertt
from loaders.MultimodalQA import MultiModalQA
from loaders.SpreadSheetQA import SpreadSheetQA

DATASET_CLASSES = {
    "WikiSQL": WikiSQL,
    "MultimodalQA": MultiModalQA,
    "WikiTableQuestions": WikiTableQuestion,
    "TATQA": TATQA,
    "FinQA": FinQA,
    "HiTab": HiTab,
    "HybridQA": HybridQA,
    "SpreadSheetQA": SpreadSheetQA,
    "MultiHiertt": MultiHiertt
}


import os
import platform

import re
import logging
import json
from structs.data import Cell, Table, HFTabularDataset


logger = logging.getLogger(__name__)

class HiTab(HFTabularDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def get_all_coordinates(self, json_data):
        coordinates = []

        def dfs(node):
            row_index = node.get("row_index", 0) 
            col_index = node.get("column_index", 0)
            coordinates.append((row_index, col_index))

            for child in node.get("children", []):
                dfs(child)

        dfs(json_data)
        if (-1,-1) in coordinates:
             coordinates.remove((-1,-1))
        return coordinates    

    def _load_split(self, split):
        question_file_map = {
            "train": "train.jsonl",
            "dev": "dev.jsonl",
            "test": "test.jsonl"
        }

        file_base_path = 'datasets/HiTab'
        logger.info(f"Loading HiTab - {split}")
        question_file_path = os.path.join(file_base_path, question_file_map[split])



        table_data_dict = {}
        for filename in os.listdir(os.path.join(file_base_path, "tables")):
            if filename.endswith(".json"):
                file_path = os.path.join(file_base_path, "tables", filename)
                with open(file_path, 'r', encoding='utf-8') as f:
                    content = json.load(f)
                    table_data_dict[filename[:-5]] = content  


        data = []

        with open(question_file_path, 'r', encoding='utf-8') as file:
            for line in file:
                question_data = json.loads(line)
                properties_info = {}

                if 'table_source' in question_data:
                    properties_info['table_source'] = str(question_data['table_source'])
                if 'aggregation' in question_data:
                    properties_info['aggregation'] = str(question_data['aggregation'])
                
                question = question_data["question"]

                table_data = table_data_dict[question_data["table_id"]]
                
                if 'title' in table_data:
                    properties_info['table_title'] = str(table_data['title'])
                if 'top_header_rows_num' in table_data:
                    properties_info['top_header_rows_num'] = str(table_data['top_header_rows_num'])
                if 'left_header_columns_num' in table_data:
                    properties_info['left_header_columns_num'] = str(table_data['left_header_columns_num'])



                table_data_text = table_data["texts"]

                col_header_idx = self.get_all_coordinates(table_data['top_root']) 
                row_header_idx = self.get_all_coordinates(table_data['left_root'])

                data.append({
                    "question": question,
                    "table": {
                        "text": table_data_text,
                        "col_header": col_header_idx,
                        "row_header": row_header_idx
                    },
                    "properties": properties_info
                })
        self.data[split] = data
        self.dataset_info = {
            "citation": "111",
            "description": "111",
            "version": "111",
            "license": "111"
        }

    def load_split_test(self, split):
        self._load_split(split)

    def prepare_table(self, entry):
        t = Table()
        t.type = "default"
        t.default_question = entry["question"]

        t.props = entry['properties']
       

        for (i,row) in enumerate(entry["table"]["text"]):
            for (j,cell) in enumerate(row):
                c = Cell()
                c.value = cell
                c.is_col_header = (i,j) in entry["table"]["col_header"]
                c.is_row_header = (i,j) in entry["table"]["row_header"]
                t.add_cell(c)
            t.save_row()

        return t
if __name__ == '__main__':
    HiTab().load_split_test("train")


from structuredqakit import ToolBaseForStructuredQA, BartForStructuredQA, BertForStructuredQA
from utils.dataset import WikisqlDataset, WikitablequestionsDataset, MSR_SQADataset, StructuredQADatasetFromBuilder_Tapas, StructuredQADatasetFromBuilder

#### BART-based model ####
# tool = BartForStructuredQA(model_name_or_path = '/home/lfy/PTM/reastap-large')
# tool = BartForStructuredQA(model_name_or_path = '/home/lfy/PTM/omnitab-large-finetuned-wtq')
# tool = BartForStructuredQA(model_name_or_path = '/home/lfy/PTM/omnitab-large-finetuned-wikisql')
# tool = BartForStructuredQA(model_name_or_path = '/home/lfy/PTM/microsoft/tapex-large-finetuned-wtq')
# tool = BartForStructuredQA(model_name_or_path = '/home/lfy/PTM/microsoft/tapex-large-finetuned-wikisql')
# tool = BartForStructuredQA(model_name_or_path = '/home/lfy/PTM/tapex-large-finetuned-sqa')
# tool = BartForStructuredQA(model_name_or_path = '/home/lfy/PTM/reastap-large-finetuned-wtq')
# tool = BartForStructuredQA(model_name_or_path = '/home/lfy/PTM/reastap-large-finetuned-wikisql')

#### BERT-based model (tapas) ####
# tool = BertForStructuredQA(model_name_or_path='/home/lfy/PTM/google/tapas-large')
# tool = BertForStructuredQA(model_name_or_path='/home/lfy/PTM/google/tapas-large-finetuned-sqa')
# tool = BertForStructuredQA(model_name_or_path='/home/lfy/PTM/google/tapas-large-finetuned-wikisql-supervised')
# tool = BertForStructuredQA(model_name_or_path='/home/lfy/PTM/google/tapas-large-finetuned-wtq')



"""select dataset"""
### BART ###
# dataset = WikisqlDataset()
# dataset = WikitablequestionsDataset()
# dataset = MSR_SQADataset()
# dataset = StructuredQADatasetFromBuilder('wikitq')
# dataset = StructuredQADatasetFromBuilder('hybridqa')

### BERT(tapas) ###
# dataset = StructuredQADatasetFromBuilder_Tapas('wikisql_tapas')
# dataset = StructuredQADatasetFromBuilder_Tapas('wikitq_tapas')
# dataset = MSR_SQADataset()


"""eval / predict"""
# tool.eval(dataset=dataset, per_device_eval_batch_size = 8)
# tool.predict(dataset= dataset)
# tool.train(dataset=dataset, per_device_train_batch_size = 4)

"""
CUDA_VISIBLE_DEVICES=0,1,2,3,4,5,6,7; python -m torch.distributed.launch --nproc_per_node 8 test.py
WANDB_DISABLED="true", CUDA_VISIBLE_DEVICES=0 python test.py
"""
# tool.train(dataset=dataset, per_device_train_batch_size = 4)


import json
import os

ROOT_DIR = os.path.join(os.path.dirname(__file__), "../")
DATASET_SPLIT = 'dev'
TYPE_FILE = os.path.join(ROOT_DIR, f"utils/retriever_files/question_type_{DATASET_SPLIT}.json")
QIM_FILE = os.path.join(ROOT_DIR, f"utils/retriever_files/question_image_match_scores_{DATASET_SPLIT}.json")
QPM_FILE = os.path.join(ROOT_DIR, f"utils/retriever_files/question_passage_match_scores_{DATASET_SPLIT}.json")

"""
Match Score Files be like:
{
    "eid": {
        "doc_id1": score,
        "doc_id2": score,
    }
}
"""


class Retriever:
    def __init__(self, type_file = TYPE_FILE, qim_file = QIM_FILE, qpm_file = QPM_FILE):
        self.type_dict = json.load(open(type_file, 'r'))
        self.qim_dict = json.load(open(qim_file, 'r'))
        self.qpm_dict = json.load(open(qpm_file, 'r'))
    
    def retrieve(self, g_data_item, eid = None, qid = None):
        if eid is None or qid is None:
            raise ValueError("eid or qid is None")

        qtype = self.type_dict[qid]['predict']
        # Get the retrieved passages
        new_passages = {"id": [], "title": [], "text": [], "url": []}
        r_passage_ids = []
        passage_scores_dict = self.qpm_dict[str(eid)].copy()
        # get top3
        if len(passage_scores_dict) > 3:
            passage_scores_dict = dict(sorted(passage_scores_dict.items(), key=lambda x:x[1], reverse=True)[:3])

        for _id, title, text, url in zip(g_data_item['passages']['id'],
                                            g_data_item['passages']['title'],
                                            g_data_item['passages']['text'],
                                            g_data_item['passages']['url']):
                if _id in r_passage_ids:
                    continue
                r_passage_ids.append(_id)
                flag = True
                # TODO: decide when set flag to False
                if not (_id in passage_scores_dict.keys()):
                    flag = False

                if flag:
                    new_passages["id"].append(_id)
                    new_passages["title"].append(title)
                    new_passages["text"].append(text)
                    new_passages["url"].append(url)
        
        # Get the retrieved images
        new_images = {"id": [], "title": [],"pic": [], "url": [], "path": []}
        r_image_ids = []
        image_scores_dict = self.qim_dict[str(eid)].copy()
        # get top3
        if len(image_scores_dict) > 3:
            image_scores_dict = dict(sorted(image_scores_dict.items(), key=lambda x:x[1], reverse=True)[:3])

        for _id, title, pic, url, path in zip(g_data_item['images']['id'],
                                                g_data_item['images']['title'],
                                                g_data_item['images']['pic'],
                                                g_data_item['images']['url'],
                                                g_data_item['images']['path']):
                if _id in r_image_ids:
                    continue
                r_image_ids.append(_id)
                flag = True
                # TODO: decide when set flag to False
                if not (_id in image_scores_dict.keys()):
                    flag = False

                if flag:
                    new_images["id"].append(_id)
                    new_images["title"].append(title)
                    new_images["pic"].append(pic)
                    new_images["url"].append(url)
                    new_images["path"].append(path)

        return new_passages, new_images, qtype

if __name__ == '__main__':
    pass